"""
 * The MIT License
 *
 * Copyright (c) 2024 Patrick Hammer
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 * """


def Prettyprint_Plan(actionlist):
    return [get_pretty_action(x) for x in actionlist[1:]]


def prettyprint_all_actions(actionlist):
    return [get_pretty_action(x) for x in actionlist]


def pp_val(val, val_to_char_mappings):
    # return the char value of the board (if there is one), rather than the internal int value
    if val in val_to_char_mappings:
        return f"'{val_to_char_mappings[val]}'"
    else:
        return f"{val}"


def prettyprint_rule(rule_evidence, Hypothesis_TruthValue, rule, val_to_char_mappings={},
                     print_evidence=False, print_truth_value=True):

    assert rule in rule_evidence

    actions_values_preconditions = rule[0]
    action = get_pretty_action(actions_values_preconditions[0])
    precons = actions_values_preconditions[2:]
    line = "<("
    line += "AgentStatePreCon=" + str(list(actions_values_preconditions[1]))+", "

    for i, x in enumerate(precons):
        line += f"{_prettyTriplet(x, val_to_char_mappings)}"
        if i != len(precons) - 1:
            line += f", "

    scoreInc = f"score+={rule[1][3][0]}"
    keys = f"AgentStatePreCon={list(rule[1][3][1:])}"
    action = action.rjust(6) # print actions with same width

    evidence = ""
    truth_value = ""
    if print_evidence:
        evidence = " evidence="+str(rule_evidence[rule])
    if print_truth_value:
        truth_value = " TruthValue="+str(Hypothesis_TruthValue(rule_evidence[rule]))

    line += (","+ action +  ")" +
             "=/> (" + keys + ", " + str(_prettyTriplet(rule[1],val_to_char_mappings)) + ", " + scoreInc + ")>."
               +truth_value+evidence)

    # print(line)
    return line


def _prettyTriplet(triplet, val_to_char_mappings):
    (r, c, value) = triplet[:3]
    if r >= 0: r = " " + str(r)
    if c >= 0: c = " " + str(c)
    str_val = pp_val(value, val_to_char_mappings)
    return "rc[" + str(r) + "," + str(c) + f" ]={str_val}"


def get_pretty_action(action):
    if str(action).find("<function ") > -1:
        return "" + str(action).split("<function ")[1].split(" at")[0]
    else:
        return "" + str(action)



"""
 * The MIT License
 *
 * Copyright (c) 2024 Patrick Hammer
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 * """

import random
import sys
from copy import deepcopy

# from numba import jit

from nace.prettyprint import prettyprint_rule


    # global DisableLRUDOpSymmetryAssumption
    # DisableLRUDOpSymmetryAssumption = True # hmmm bad double negative



# @jit(nopython=True)
def Hypothesis_TruthValue(wpn):
    """
    # The truth value of a hypothesis can be obtained directly from the positive and negative evidence counter

    @param wpn:
    @return:
    """
    (wp, wn) = wpn
    frequency = wp / (wp + wn)
    confidence = (wp + wn) / (wp + wn + 1)
    return frequency, confidence


# @jit(nopython=True)
def _TruthExpectation(tv):
    """
    # The truth expectation calculation based on the truth value (frequency, confidence) tuple

    @param tv: truth value made up from frequency and confidence
    @return:
    """
    (f, c) = tv
    return c * (f - 0.5) + 0.5


def _Choice(rule_evidence, rule1, rule2):
    """
    # When two hypotheses predict a different outcome for the same conditions, the higher truth exp one is chosen

    @param rule_evidence:
    @param rule1:
    @param rule2:
    @return:
    """
    if rule1 in rule_evidence and rule2 not in rule_evidence:  # dv added (no longer triggered)
        return rule1
    if rule1 not in rule_evidence and rule2 in rule_evidence:  # dv added (no longer triggered)
        return rule2
    if rule1 not in rule_evidence and rule2 not in rule_evidence:  # dv added (no longer triggered)
        # panic
        print("ERROR this suggests a logic error in calling code.")
        return rule2

    t1 = Hypothesis_TruthValue(rule_evidence[rule1])
    t2 = Hypothesis_TruthValue(rule_evidence[rule2])
    if _TruthExpectation(t1) > _TruthExpectation(t2):
        return rule1
    return rule2


def Hypothesis_Contradicted(rule_evidence, ruleset, negruleset, rule, val_to_char_mappings):  # this mutates returned copy of rule_evidence
    """
    # Negative evidence was found for the hypothesis/rule

    @param rule_evidence:
    @param ruleset:
    @param negruleset:
    @param rule:
    @return:
    """
    rule_evidence = _AddEvidence(rule_evidence, rule, False)  # mutates returned copy of rule_evidence
    if "silent" not in sys.argv:
        print("Neg. revised: ", end="")
        line = prettyprint_rule(rule_evidence, Hypothesis_TruthValue, rule, val_to_char_mappings)
        print(line)
        # in a deterministic setting this would have sufficed however
        # simply excluding rules does not work in non-deterministic ones
        # if rule in ruleset:
        #    print("RULE REMOVAL: ", end=""); Prettyprint_rule(rule_evidence, Hypothesis_TruthValue, rule)
        #    ruleset.remove(rule)
        # negruleset.add(rule)
    return rule_evidence, ruleset, negruleset


def Hypothesis_Confirmed(  # this mutates the returned rule_evidence and ruleset
        FocusSet, rule_evidence, ruleset, negruleset, rule, val_to_char_mappings:dict
):
    """
    # Positive evidence was found for the hypothesis/rule
    # Confirm rule against +ve and -ve evidence, add variants (i.e. transforms) to newrules.

    @param FocusSet:
    @param rule_evidence: dict[rule:(+ve_evidence, -ve_evidence)]
    @param ruleset:
    @param negruleset:
    @param rule:
    @return:
    """

    rule_evidence = deepcopy(rule_evidence)  # dv added this line
    # try location symmetry
    variants = [rule]

    for i, r in enumerate(variants):
        if i > 0:  # abduced hypotheses
            if r in rule_evidence:  # this derived hypothesis already exists
                continue
        rule_evidence = _AddEvidence(rule_evidence, r, True)  # mutates returned copy of rule_evidence
        if "silent" not in sys.argv:
            print("Pos. revised: ", end="")
            line = prettyprint_rule(rule_evidence, Hypothesis_TruthValue, r, val_to_char_mappings)
            print(line)
        if r not in negruleset:
            if r not in ruleset:
                # print("RULE ADDITION: ", end=""); Prettyprint_rule(rule)
                ruleset.add(r)
    return rule_evidence, ruleset


def Hypothesis_ValidCondition(cond):
    """
    # Valid condition predicate defining the accepted neighbourhood between conclusion and condition cells
    # restrict to neighbours (CA assumption)
    # If 0,1, or 2 in distance in any direction, return True, else False

    @param cond:
    @return:
    """
    (y, x, v) = cond
    if y == 0 and x == 0:  # self
        return True
    if y == 0 and (x == -1 or x == -2):  # left
        return True
    if (y == -1 or y == -2) and x == 0:  # up
        return True
    if y == 0 and (x == 1 or x == 2):  # right
        return True
    if (y == 1 or y == 2) and x == 0:  # down
        return True

    # Does this break the CA assumption? what is the CA assumption?

    if abs(y) == 1 and abs(x) == 1:
        # diagonal is being allowed, dec 2024, for deepsea and frozen lake.
        return True

    return False


def Hypothesis_BestSelection(rules, rules_excluded, rule_evidence,
                             include_random_rules):  # mutates returned rules, returned rulesExcluded
    """

    We exclude rules which have more negative evidence than positive, and choose the highest truth-exp ones whenever
    a different outcome would be predicted for the same conditions

    @param rules:
    @param rules_excluded:
    @param rule_evidence:
    @param include_random_rules: World stayed the same when last ground truth copied in - if true stochastic element added
    @return:
    """
    rules = deepcopy(rules)    # ensure we don't mutated passed in data structures.
    rules_excluded = deepcopy(rules_excluded)
    rule_evidence = deepcopy(rule_evidence)


    rulesin = deepcopy(rules)
    rules_moved_count_1 = 0
    rules_moved_count_2 = 0
    rules_moved_count_3 = 0
    for i, rule1 in enumerate(rulesin):
        # if Hypothesis_TruthExpectation(Hypothesis_TruthValue(rule_evidence[rule1])) <= 0.5: #exclude rules which
        # are not better than exp (only 0.5+ makes sense here)
        if rule1 in rule_evidence:  # dv 8/jul/24 added to stop key not found which hadn't happened before
            te = _TruthExpectation(Hypothesis_TruthValue(rule_evidence[rule1]))
            if te <= 0.5 or (
                    include_random_rules
                    and random.random()
                    > _TruthExpectation(Hypothesis_TruthValue(rule_evidence[rule1]))
            ):
                if rule1 in rules:
                    rules_excluded.add(rule1)
                    rules.remove(rule1)
                    rules_moved_count_1 += 1
    rulesin = deepcopy(rules)
    for i, rule1 in enumerate(rulesin):
        for j, rule2 in enumerate(rulesin):
            if (
                    i != j
            ):  # exclude rules of same precondition which are worse by truth value
                if rule1[0] == rule2[0]:
                    rulex = _Choice(rule_evidence, rule1, rule2)
                    if rulex == rule1:
                        if rule2 in rules:
                            rules_excluded.add(rule2)
                            rules.remove(rule2)
                            rules_moved_count_2 += 1
                            # print("excluded ", end=''); Prettyprint_rule(rule2)
                    else: # best rule is rule2, so remove rule1
                        if rule1 in rules:
                            rules_excluded.add(rule1)
                            rules.remove(rule1)
                            rules_moved_count_3 += 1
                            # print("excluded", end=''); Prettyprint_rule(rule1)
    print("DEBUG Rules excluded for predictions:  rules_less_than_truth_value_less_than_half=", rules_moved_count_1, "same_precondition_worse_truth_value=", (rules_moved_count_2+ rules_moved_count_3), "active_rules=", len(rules) )

    return rules, rules_excluded


def _AddEvidence(rule_evidence, rule, positive, w_max=20):  # Mutates a copy of rule_evidence
    """
    Add positive or negative evidence for a rule, with a certain max. amount of evidence so that non-stationary
    environments can be handled too

    @param rule_evidence:
    @param rule:
    @param positive:
    @param w_max:
    @return:
    """
    rule_evidence = deepcopy(rule_evidence)
    if rule not in rule_evidence:
        rule_evidence[rule] = (0, 0)
    (wp, wn) = rule_evidence[rule]
    if positive:
        if wp + wn <= w_max:
            rule_evidence[rule] = (wp + 1, wn)
        else:
            rule_evidence[rule] = (wp, max(0, wn - 1))
    else:
        if wp + wn <= w_max:
            rule_evidence[rule] = (wp, wn + 1)
        else:
            rule_evidence[rule] = (max(0, wp - 1), wn)
    return rule_evidence



"""

utilities to aid in test classes.

"""
import copy
import numpy as np

import nace
# from nace.nace_v3 import _get_rc_delta_for_action_list

def convert_focus_set_to_internal_mapping(focus_set, world):
    result = {}
    for ch in focus_set.keys():
        val = world.get_val_for_char(ch)
        result[val] = focus_set[ch]
    return result

def convert_focus_set_to_char_mapping(focus_set, world):
    # defining things in chars is nice for unit tests, but useless in practice as too few states can be held.
    # give a way to convert rules that are used in prod into a nice unit test version.

    mappings = world.get_val_to_char_mappings(list_of_extra_values = [ k for k in focus_set.keys()])
    result = {}
    for val in focus_set.keys():
        if val in mappings:
            ch = mappings[val]
        else:
            print("ERROR R26")

        result[ch] = focus_set[val]
    return result


def convert_rules_to_internal_mapping(rules, world):
    result = set()
    for rule in rules:
        precondition, effect = rule
        new_precondition = list(precondition)
        new_effect = list(effect)
        new_effect[2] = world.get_val_for_char(new_effect[2])
        new_effect = tuple(new_effect)
        for outer_index, element in enumerate(new_precondition):
            if isinstance(element, tuple):
                if len(element) == 3:
                    element = list(element)
                    element[2] = world.get_val_for_char(element[2])
                    new_precondition[outer_index] = tuple(element)
        result.add( (tuple(new_precondition), new_effect )  )
    return result

def convert_rule_evidence_to_internal_mapping(rule_evidence, world):
    result = {}
    for rule in rule_evidence.keys():
        precondition, effect = rule
        new_precondition = list(precondition)
        new_effect = list(effect)
        new_effect[2] = world.get_val_for_char(new_effect[2])
        new_effect = tuple(new_effect)
        for outer_index, element in enumerate(new_precondition):
            if isinstance(element, tuple):
                if len(element) == 3:
                    element = list(element)
                    element[2] = world.get_val_for_char(element[2])
                    new_precondition[outer_index] = tuple(element)
        new_rule = tuple((tuple(new_precondition), new_effect ))
        result[new_rule] = rule_evidence[rule]
    return result


def convert_rule_evidence_to_char_or_int_mappings(rule_evidence, world, dest="CHAR"):
    char_rule_evidence = {}
    for rule in rule_evidence.keys():
        t_char_rules, t_encoded_rules = convert_rules_to_char_or_int_mappings([rule], world, dest=dest)
        char_rule_evidence[t_encoded_rules[0]] = rule_evidence[rule]
    return char_rule_evidence


def convert_rules_to_char_or_int_mappings(rules, world, dest="CHAR"):
    # defining things in chars is nice for unit tests, but useless in practice as too few states can be held.
    # give a way to convert rules that are used in prod into a nice unit test version.


    # NOTE the row and column in the effect are always zero and not currently used.

    list_of_extra_values = [rule[1][2] for rule in rules if rule is not None]
    mappings = world.get_val_to_char_mappings(list_of_extra_values)
    initial_results = []
    values_to_order_by = []
    for rule in rules:
        precondition, effect = rule
        new_precondition = list(precondition)
        new_precondition[0] = nace.prettyprint.get_pretty_action(new_precondition[0])
        new_effect = list(effect)
        if dest == "CHAR":
            if new_effect[2] in mappings:
                new_effect[2] = mappings[new_effect[2]]
            else:
                pass # we can not decode the item, leave it in its raw state.
        else:
            if isinstance(new_effect[2], int):
                pass # already an int
            else:
                new_effect[2] = new_effect[2].item()
        for outer_index, element in enumerate(new_precondition):
            if isinstance(element, tuple):
                if len(element) == 3:
                    if dest == "CHAR":
                        element = list(element)
                        if element[2] in mappings:
                            element[2] = mappings[element[2]]
                            values_to_order_by.append(element[2])
                        else:
                            mappings2 = world.get_val_to_char_mappings(list_of_extra_values=[element[2]])
                            if element[2] in mappings2:
                                element[2] = mappings2[element[2]]
                                values_to_order_by.append(element[2])
                            else:
                                _ = world.get_val_for_char(element[2])
                                print("WARN: error converting value to char, being left as is.", element[2])
                    else:
                        # INT
                        element = list(element)
                        if isinstance(element[2], int):
                            pass # already an int
                        else:
                            element[2] = element[2].item()
                        values_to_order_by.append(element[2])
                    new_precondition[outer_index] = tuple(element)
        initial_results.append( (tuple(new_precondition), tuple(new_effect) )  )

    # get an ordered set of all the chars at 0,0 in each rule
    values_to_order_by = list(set(values_to_order_by))
    for i, result in enumerate(copy.deepcopy(initial_results)):
        precondition, effect = result
        for condition in precondition[2:]:
            if condition[0] == 0 and condition[1] == 0:
                values_to_order_by.append(condition[2])
    values_to_order_by = list(set(values_to_order_by))
    try:
        values_to_order_by.sort()
    except TypeError as e:
        print("WARN trying to sort str and int caused an error which has been supressed.")
        pass # had mixed typed in the list. this is a temp work around.

    # order the rules (for ease of debugging)
    ordered_results = []
    order_results_as_str = "[\n"
    for value in values_to_order_by:
        not_yet_commented = True
        for i, result in enumerate(copy.deepcopy(initial_results)):
            precondition, effect = result
            matches = False
            for condition in precondition[2:]:
                if condition[0] == 0 and condition[1] == 0 and condition[2] == value:
                    matches = True
            if matches:
                ordered_results.append(result)
                order_results_as_str += str(result)+","
                if not_yet_commented:
                    not_yet_commented = False
                    order_results_as_str += " # "+str(value)
                order_results_as_str += "\n"

        for result in ordered_results:
            if result in initial_results:
                initial_results.remove(result)

    order_results_as_str += "]"
    assert len(initial_results) == 0

    return order_results_as_str, ordered_results






def get_xy_delta_for_action_list(action_list):
    # DO NOT call this routine, we are moving to r,c everywhere
    # ignores walls etc.
    print("WARN: MOVE to rc, do not use xy.")
    r,c = get_rc_delta_for_action_list(action_list)
    return (c,r)

def get_rc_delta_for_action_list(action_list):
    # ignores walls etc. only good for test code
    return nace.nace_v3._get_rc_delta_for_action_list(action_list)

def get_time_and_board_at_destination(list_of_xy: list, world: nace.world_module_numpy.NPWorld):
    x = sum([r[0] for r in list_of_xy])
    y = sum([r[1] for r in list_of_xy])
    return world.times[y][x], world.board[y][x]




import copy
import glob
from collections import defaultdict

import gymnasium as gym
import numpy as np

import nace

"""

Inprogress: code to find movement action effects from actions, 
i.e. was trying to not need to know the meaning of 'up', but learn it from experience.

"""


def perform_actions_find_location_deltas(env_name, cell_shape_rc, action_number_list=(1,)):
    """
    Perform all the actions in the list on a newly initialised world. Return the agent location.

    :param env_name:
    :param action_number_list:
    :return:
    """
    # check the agents location is where it is expected.
    agent_indication_raw_value_list = []
    fn_list = glob.glob("./data/agents/*.npy")
    for fn in fn_list:
        cell = np.load(fn)
        agent_indication_raw_value_list.append(cell)
    env = gym.make(env_name, render_mode="rgb_array")
    env.reset()
    env_rendered = env.render()
    external_ground_truth_npworld = nace.world_module_numpy.NPWorld(with_observed_time=False,
                                                                    name="external ground truth")
    original_agent_location, _, _, _ = external_ground_truth_npworld.update_world_from_ground_truth_NPArray(
        env_rendered,
        update_mode='ALL',
        cell_shape_rc=cell_shape_rc,
        add_surrounding_walls=True,
        wall_code=nace.world_module_numpy.EMBEDDED_WALL_CODE,
        agent_indication_raw_value_list=agent_indication_raw_value_list,
        observed_at=1
    )
    if len(original_agent_location) == 0:
        raise Exception("Could not find agent")

    original_agent_location = original_agent_location[-1]
    locations = [original_agent_location]
    embedded_values = [
        external_ground_truth_npworld.get_embedded_val_rc(original_agent_location[0], original_agent_location[1])]
    for observed_at, action_number in enumerate(action_number_list):
        env.step(action_number)  # 0 == up?
        env_rendered = env.render()
        external_ground_truth_npworld = nace.world_module_numpy.NPWorld(with_observed_time=False,
                                                                        name="external ground truth")
        _, _, _, _ = (external_ground_truth_npworld.update_world_from_ground_truth_NPArray(
            env_rendered,
            update_mode='ALL',
            cell_shape_rc=cell_shape_rc,
            add_surrounding_walls=True,
            wall_code=nace.world_module_numpy.EMBEDDED_WALL_CODE,
            agent_indication_raw_value_list=agent_indication_raw_value_list,
            observed_at=observed_at
        ))
        location = external_ground_truth_npworld.extract_agent_location_raw(agent_indication_raw_value_list)
        locations.append(location)
        embedded_value = external_ground_truth_npworld.get_embedded_val_rc(location[0], location[1])
        embedded_values.append(embedded_value)

    if location == (-1, -1):
        return None, False
    if len(original_agent_location) > 0 and location == original_agent_location[0]:
        return None, False

    last_move_delta = (locations[-1][0] - locations[-2][0], locations[-1][1] - locations[-2][1])

    embedding_change = (embedded_values[-2], embedded_values[-1])

    return last_move_delta, embedding_change


def find_gym_action_mappings(env_name="CliffWalking-v0", num_actions=4, cell_shape_rc=(60, 60)):
    """
    determine which actions are up down left right.
    :return:
    """
    results1 = {}
    embedding_change_by_action1 = {}
    for action_number_1 in range(num_actions):
        movement_delta, agent_embedding_change = perform_actions_find_location_deltas(env_name=env_name,
                                                                                      cell_shape_rc=cell_shape_rc,
                                                                                      action_number_list=[
                                                                                          action_number_1])
        if movement_delta is not None:
            if not (movement_delta[0] == 0 and movement_delta[1] == 0):
                if abs(movement_delta[0]) + abs(movement_delta[1]) == 1:
                    results1[action_number_1] = movement_delta
            if movement_delta == (0, 0) and agent_embedding_change[0] != agent_embedding_change[1]:
                if action_number_1 not in embedding_change_by_action1:
                    embedding_change_by_action1[action_number_1] = []
                embedding_change_by_action1[action_number_1].append(agent_embedding_change)

    results2 = {}
    embedding_change_by_action2 = {}
    for action_number_1 in results1.keys():
        for action_number_2 in range(num_actions):
            movement_delta, agent_embedding_change = perform_actions_find_location_deltas(env_name="CliffWalking-v0",
                                                                                          cell_shape_rc=cell_shape_rc,
                                                                                          action_number_list=[
                                                                                              action_number_1,
                                                                                              action_number_2])
            if movement_delta is not None:
                if not (movement_delta[0] == 0 and movement_delta[1] == 0):
                    if abs(movement_delta[0]) + abs(movement_delta[1]) == 1:
                        results2[action_number_2] = movement_delta
                if movement_delta == (0, 0) and agent_embedding_change[0] != agent_embedding_change[1]:
                    if action_number_2 not in embedding_change_by_action2:
                        embedding_change_by_action2[action_number_2] = []
                    embedding_change_by_action2[action_number_2].append(agent_embedding_change)

    results3 = {}
    embedding_change_by_action3 = {}
    for action_number_1 in results1.keys():
        for action_number_2 in results2.keys():
            for action_number_3 in range(num_actions):
                movement_delta, agent_embedding_change = perform_actions_find_location_deltas(
                    env_name="CliffWalking-v0", cell_shape_rc=cell_shape_rc,
                    action_number_list=[action_number_1, action_number_2, action_number_3])
                if movement_delta is not None:
                    if not (movement_delta[0] == 0 and movement_delta[1] == 0):
                        if abs(movement_delta[0]) + abs(movement_delta[1]) == 1:
                            results3[action_number_3] = movement_delta
                if movement_delta == (0, 0) and agent_embedding_change[0] != agent_embedding_change[1]:
                    if action_number_3 not in embedding_change_by_action3:
                        embedding_change_by_action3[action_number_3] = []
                    embedding_change_by_action3[action_number_3].append(agent_embedding_change)

    results4 = {}
    embedding_change_by_action4 = {}
    for action_number_1 in results1.keys():
        for action_number_2 in results2.keys():
            for action_number_3 in results3.keys():
                for action_number_4 in range(num_actions):
                    movement_delta, agent_embedding_change = perform_actions_find_location_deltas(
                        env_name="CliffWalking-v0", cell_shape_rc=cell_shape_rc,
                        action_number_list=[action_number_1, action_number_2, action_number_3, action_number_4])
                    if movement_delta is not None:
                        if not (movement_delta[0] == 0 and movement_delta[1] == 0):
                            if abs(movement_delta[0]) + abs(movement_delta[1]) == 1:
                                results4[action_number_4] = movement_delta
                    if movement_delta == (0, 0) and agent_embedding_change[0] != agent_embedding_change[1]:
                        if action_number_4 not in embedding_change_by_action4:
                            embedding_change_by_action4[action_number_4] = []
                        embedding_change_by_action4[action_number_4].append(agent_embedding_change)

    # print( json.dumps(results, indent=2))
    # print( json.dumps(results2, indent=2))
    # print( json.dumps(results3, indent=2))
    # print( json.dumps(results4, indent=2))
    #  we expect {0:nace.world_module.up, 1:nace.world_module.right, 2:nace.world_module.down, 3:nace.world_module.left}
    #                 -1,0                     0,1                        1,0                       0,-1

    all_deltas = {}
    all_deltas.update(results1)
    all_deltas.update(results2)
    all_deltas.update(results3)
    all_deltas.update(results4)
    # print( json.dumps(all, indent=2))

    agent_embedding_change = {}
    for d in [embedding_change_by_action1, embedding_change_by_action2, embedding_change_by_action3,
              embedding_change_by_action4]:
        for k in d.keys():
            if k not in agent_embedding_change:
                agent_embedding_change[k] = []
            agent_embedding_change[k].extend(d[k])

    # reformat into which number do which actions
    movement_action_list = [None] * num_actions
    for k in all_deltas.keys():
        if all_deltas[k] == (-1, 0):
            movement_action_list[k] = nace.world_module.up
        if all_deltas[k] == (1, 0):
            movement_action_list[k] = nace.world_module.down
        if all_deltas[k] == (0, -1):
            movement_action_list[k] = nace.world_module.left
        if all_deltas[k] == (0, 1):
            movement_action_list[k] = nace.world_module.right

    # movement_action_list : list, the left, right, up, down action occuring in each location in the coresponing to the gym_action_number
    # agent_embedding_change for each gym_action_number, i.e. for action 0, the agent embedding might change efrom 1 to 2 0:(1,2)
    return movement_action_list, agent_embedding_change


def perform_actions_find_observation_deltas(env_name, action_number_list=[1]):
    """
    Perform all the actions in the list on a newly initialised env. find changes in the observations.

    :param env_name:
    :param action_number_list:
    :return:
    """

    env = gym.make(env_name, render_mode="rgb_array")
    observation, info = env.reset()
    list_of_observations = [observation]
    env_rendered = env.render()

    observation_deltas = {}

    for action in action_number_list:
        post_action_observation, reward, terminated, truncated, info = env.step(action)
        list_of_observations.append(post_action_observation)

        if action not in observation_deltas:
            observation_deltas[action] = []

        observation_deltas[action].append(list_of_observations[-1] - list_of_observations[-2])

    return observation_deltas


def _find_action_observation_relationships(node, env_name="CliffWalking-v0", num_actions=4, action_prefixes=[]):
    for action in range(num_actions):
        action_list = copy.deepcopy(action_prefixes)
        action_list.append(action)

        x = perform_actions_find_observation_deltas(env_name, action_list)
        if x[action][-1] != 0:
            if action in node:
                node[action]["deltas"].append(x[action][-1])
            else:
                result = {"next": {}, "deltas": [x[action][-1]]}
                node[action] = result

    print()


def find_action_observation_relationships(env_name="CliffWalking-v0", num_actions=4):
    """
    Do a small amount of structured exploration, to get an idea of which actions affect which parts of the observation space.
    @param env_name:
    @param num_actions:
    @return:
    """
    parent_node = {}
    _find_action_observation_relationships(parent_node, env_name=env_name, num_actions=num_actions, action_prefixes=[])

    a1 = ""
    for a0 in parent_node.keys():
        child_node = parent_node[a0]["next"]
        _find_action_observation_relationships(child_node, env_name=env_name, num_actions=num_actions,
                                               action_prefixes=[a0])

        for a1 in child_node.keys():
            c1_node = child_node[a1]["next"]
            _find_action_observation_relationships(c1_node, env_name=env_name, num_actions=num_actions,
                                                   action_prefixes=[a0, a1])

    deltas_sums = {}
    delta_counts = defaultdict(int)
    for a0 in parent_node.keys():
        delta_counts[a0] += 1
        if a1 not in deltas_sums:
            deltas_sums[a0] = parent_node[a0]["deltas"]
        else:
            for i in range(len(deltas_sums[a0])):
                deltas_sums[a0][i] += parent_node[a0]["deltas"][i]

        for a1 in parent_node[a0]["next"].keys():
            delta_counts[a1] += 1
            if a1 not in deltas_sums:
                deltas_sums[a1] = parent_node[a0]["next"][a1]["deltas"]
            else:
                for i in range(len(deltas_sums[a0])):
                    deltas_sums[a1][i] += parent_node[a0]["next"][a1]["deltas"][i]

            for a2 in parent_node[a0]["next"][a1]["next"].keys():
                delta_counts[a2] += 1
                if a2 not in deltas_sums:
                    deltas_sums[a2] = parent_node[a0]["next"][a1]["next"][a2]["deltas"]
                else:
                    for i in range(len(deltas_sums[a0])):
                        deltas_sums[a2][i] += parent_node[a0]["next"][a1]["next"][a2]["deltas"][i]

    averages = {}
    for action in delta_counts.keys():
        averages[action] = copy.deepcopy(deltas_sums[action])
        for i in range(len(averages[action])):
            averages[action][i] = averages[action][i] / delta_counts[action]

    # It should be possible to tell from the averages which action is the inverse of which.
    # It might be possible to tell that action a,b and inverse, and symmetrical to actions c,d
    # if 1) they affect the same observation value with some scale factor , or
    # 2) they affect different (and possibly orthogonal) observations with some scale factor.

    """
    Return average observation change for each action.
    """
    return averages



"""
 * The MIT License
 *
 * Copyright (c) 2024 Dwane van der Sluis
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 * """
import copy
import json
import sys
import time
import random
import numpy as np
import collections
from typing import Type

import nace.world_module_numpy
from nace.agent_module import Agent
from nace.nace_v3 import nacev3_get_next_action, nacev3_predict_and_observe, print_world_at_plan_end
from nace.world_module_numpy import NPWorld
from nace.graph_utilities import produce_charts
from nace.rule_evidence_utils import load_rule_evidence_from_disk

# def _get_expected_deltas(nace_action_name):
#     expected_directions = {
#         "^left": (-1, 0),
#         "^right": (1, 0),
#         "^up": (0, -1),
#         "^down": (0, 1)
#     }
#     expected_column_delta, expected_row_delta = expected_directions.get(nace_action_name, (0, 0))
#     return expected_row_delta, expected_column_delta


class StepperV4():
    def __init__(self, agent_indication_raw_value_list:list, seed_value=1, unobserved_code=nace.world_module_numpy.UNOBSERVED_BOARD_VALUE, available_actions_function=None):
        # The world is held externally.
        # The state of the agent (score number of keys held etc.) is held externally
        # Actions are performed on that world externally
        # this class wraps code and expectations that allow it to take one of the passed in worlds and predict forward.
        # partial observation (if it occurs) happens outside this code.

        # This module hole

        # variables set to specific values and passed in
        self.current_behavior = "BABBLE"
        random.seed(seed_value)

        # variables set to empty sets, or lists before the first call of this routine

        # these 4 var are global in the original implementation
        self.focus_set = dict([])
        self.rule_evidence = dict([])
        self.rules = set()
        self.negrules = set()
        self.whole_plan = []
        self.time_counter = -1
        self.unobserved_code = unobserved_code  # indicates we have not seen the maps true value
        self.agent_indication_raw_value_list = agent_indication_raw_value_list # raw value(s) of an agent, i.e rgb np.ndarrays
        self.step_number = 0 # increments each time get_next_Action() is called. Similar, but different to time.
        self.current_plan = [] # plan currently being executed.
        self.rewards = [] # unbounded (warning could grow forever) list of rewards.

        self.anamestr = "1234"

        # variables NOT set before the first call of this routine (in original implementation)
        self.used_rules = None
        self.debuginput = None

        self.rules_excluded = None
        self.stayed_the_same = False  # when last time the new ground truth is copied over it.

        self.post_action_agent = None
        self.pre_action_agent = None

        self.internal_world = None  # working copy

        self.internal_preaction_world = None  # updated each time a best action is predicted
        self.action = None  # predicted best action
        self.available_actions_function = available_actions_function # Will be used to get the available actions
        self.statistics = collections.defaultdict(int)
        self.series_statistics = []

        self.unittest_code_str = ""

    def set_agent_ground_truth_state(self, rc_loc, score, terminated, values_exc_prefix):

        if self.post_action_agent is None:
            self.pre_action_agent = Agent(rc_loc, score, int(terminated), [0 for _ in values_exc_prefix])
            self.post_action_agent = Agent(rc_loc, score,
                                           terminated=int(terminated),
                                           values_excluding_prefix=values_exc_prefix)
        else:
            reward = score - self.post_action_agent.get_score()
            self.rewards.append(reward) # TODO unbounded, will grow forever, also currently unused.
            print("DEBUG rewards max=",max(self.rewards), "min=", min(self.rewards), "mean=", sum(self.rewards)/len(self.rewards))

            self.pre_action_agent = copy.deepcopy(self.post_action_agent)
            self.post_action_agent.set_rc_loc(rc_loc)
            self.post_action_agent.set_values_inc_prefix([score] + [int(terminated)] + list(values_exc_prefix))
        # Possible feature: try and reverse engineer rules that change each value into the description for that value
        # 'kx',left -> 'x ', v[1]+=1   --> pick up k
        # result = {}
        # for i, v in enumerate(self.post_action_agent.values):
        #     if i == Agent.INDEX_SCORE:
        #         result["score"] = {"v": v}
        #     elif i == Agent.INDEX_TERMINATED:
        #         result["terminated"] = {"v": v}
        #     else:
        #         result["v" + str(i)] = {"v": v}
        # return result

    def set_world_ground_truth_state(
            self,
            ground_truth_external_world: Type[NPWorld],
            view_rc_locs: list,
            time_counter:int):
        # Update the world in the region of the agent's field of view
        if ground_truth_external_world is not None:
            modified_count, _ = self.internal_world.update_world_from_ground_truth(
                time_counter,
                ground_truth_external_world,
                rc_locations=view_rc_locs,
            agent_indication_raw_value_list=self.agent_indication_raw_value_list)  # called 3rd Nov


    def get_next_action(self,
                        ground_truth_external_world: Type[NPWorld],
                        new_rc_loc,
                        print_debug_info,
                        time_delay_sec=0.0, view_dist_x=3, view_dist_y=2,
                        max_num_actions: int=12,  # planning search depth
                        max_queue_length: int=70,  # planning queue depth
                        use_sticky_plans:bool =True
                        ):
        self.step_number += 1
        if self.pre_action_agent is None:
            # self.pre_action_agent = Agent(new_rc_loc, score=0, values_exc_score=(0,))
            self.pre_action_agent = Agent.new_default_instance(new_rc_loc)
        if self.internal_world is None:
            self.internal_world = NPWorld(with_observed_time=True, name="self.internal_world", view_dist_x=view_dist_x,
                                          view_dist_y=view_dist_y,
                                          store_raw_cell_data=False # 5th Feb 2024
                                          )


        # new_rc_loc be the single location of the agent we are interested in (if multiple)
        assert isinstance(new_rc_loc, tuple)

        start_time = time.time()
        self.time_counter += 1
        agent = self.post_action_agent if self.post_action_agent is not None else self.pre_action_agent

        if self.available_actions_function is not None:
            full_action_list = self.available_actions_function()
        else:
            raise Exception("Pass the function that returns valid actions to the Stepper constructor.")

        # if we use the difference count, we then want to use which differences matter and which don't,
        # which seems like a step to far without being extremely general.
        differences_count = self.internal_world.get_difference_count(self.internal_preaction_world)
        print("DEBUG world difference count (t-1, and now) =", differences_count,
              " Note: this should never be used in our logic as it it too environment dependent.")

        if (len(self.whole_plan) > 1 and use_sticky_plans):
            # we are executing a previous planed plan,  remove the action just executed. Set the next action.
            self.whole_plan = self.whole_plan[1:]
            self.action = self.whole_plan[0]
            statistics = {}
            print("INFO Using sticky plan ", self.whole_plan)
        else:
            self.whole_plan, self.rules_excluded, self.current_behavior, statistics, unittest_code_str = nacev3_get_next_action(
                self.time_counter,
                self.focus_set,
                self.rule_evidence,  # can be mutated
                [agent.get_rc_loc()],  #
                self.internal_world,  # passed in, has updates from external world applied to it.
                self.rules,
                ground_truth_external_world,  # used to update internal world
                print_debug_info=print_debug_info,
                stayed_the_same=self.stayed_the_same,
                agent=agent,
                full_action_list=full_action_list,
                agent_indication_raw_value_list=self.agent_indication_raw_value_list, # called 3rd Nov
                max_num_actions=max_num_actions, # planning search depth
                max_queue_length=max_queue_length # planning queue depth
            )
            self.action = self.whole_plan[0]
            # store plan
            self.unittest_code_str = unittest_code_str
            for k in statistics.keys():
                self.statistics["nacev3_get_next_action." + k] += statistics[k]
            self.statistics["plan.lowest_AIRIS_confidence"] = statistics["plan.lowest_AIRIS_confidence"]

        self.statistics["nacev3_get_next_action.call_count"] += 1
        self.statistics["plan.behavior.curious"] = 1 if self.current_behavior == "CURIOUS" else 0
        self.statistics["plan.behavior.achieve"] = 1 if self.current_behavior == "ACHIEVE" else 0
        self.statistics["plan.behavior.babble"] = 1 if self.current_behavior == "BABBLE" else 0
        self.statistics["plan.behavior.explore"] = 1 if self.current_behavior == "EXPLORE" else 0
        assert self.current_behavior in ["EXPLORE", "BABBLE", "ACHIEVE", "CURIOUS"]

        series_record = {}
        for k in ["plan.evaluation_count",
                  "plan.total_items_placed_on_queue",
                  "plan.short_circuit_advantage",
                  "plan.evaluation_count_at_which_reward_found",
                  "plan.evaluation_count_at_which_lower_AIRIS_confidence_found",
                  "plan.lowest_AIRIS_confidence",
                  "plan.behavior.achieve",
                  "plan.behavior.babble",
                  "plan.behavior.explore",
                  "plan.behavior.curious",
                  ]:
            v = self.statistics[k] if k in self.statistics else 0
            series_record[k] = v


        self.series_statistics.append(series_record)
        with open("./data/series_statistics.json", "w") as f:
            json.dump(self.series_statistics, f)

        # Store a copy after the external updates are copied in, i.e. the world use to predict on
        self.internal_preaction_world = copy.deepcopy(self.internal_world) # used in predict and observe

        end_time = time.time()
        if "manual" in sys.argv:
            print(self.current_behavior)
        else:
            if print_debug_info:
                from nace.test_utilities import (convert_focus_set_to_char_mapping)
                new_focus_set = convert_focus_set_to_char_mapping(self.focus_set, ground_truth_external_world)
                print("DEBUG get_next_action()", "next_action=", str(self.action),  self.current_behavior, "whole plan=", self.whole_plan)
                # print("INFO get_next_action()", "series_statistics=", self.series_statistics )

        elapsed_time = end_time - start_time
        if (
                elapsed_time < time_delay_sec
                and "nosleep" not in sys.argv
                and "debug" not in sys.argv
                and "manual" not in sys.argv
        ):
            time.sleep(time_delay_sec - elapsed_time)

        return self.action, self.current_behavior

    def overide_best_action(self, action):
        available_actions = self.available_actions_function()
        assert action in available_actions
        print("WARN the action has been overridden from the planned action from ",self.action," to", action)
        self.action = action

    def add_to_series_record(self, dict_of_values, prefix="", create_charts=False ):
        for k in dict_of_values.keys():
            self.series_statistics[-1][k] = dict_of_values[k]

        if create_charts:
            produce_charts(self.series_statistics, prefix)

        save_as = "./data/" + prefix + "_series_data.json"
        with open(save_as, "w") as f:
            json.dump(self.series_statistics, f)


    def predict_and_observe(self,
                            object_count_threshold=1,
                            print_out_world_and_plan=True
                            ):
        """
        Updates the KB of rules and evidence based on the differences between the last state, the predicted state,
        and the actual new state of the environment, and the world.

        Stores as many of the variables in the 'Stepper' object so calling code is simplier.

        @param object_count_threshold: how many new instances can be seen , and be considered, more are ignored
        @return:
        """
        (
            self.used_rules,
            self.focus_set,
            self.rule_evidence,
            predicted_world,
            self.rules,
            self.negrules,
            values,
            _lastplanworld,
            _planworld,
            self.stayed_the_same,
            stats
        ) = nacev3_predict_and_observe(
            self.time_counter,
            self.focus_set,
            self.rule_evidence,
            self.pre_action_agent.get_rc_loc(),
            copy.deepcopy(self.internal_preaction_world),  # pre action world (with times)
            self.rules,  # used in part1 and part2
            self.negrules,
            self.action,
            self.rules_excluded,
            copy.deepcopy(self.internal_world),
            # post action world, can be internal or external, differences are copied in
            pre_action_agent=self.pre_action_agent,  # pre action agent
            ground_truth_post_action_agent=self.post_action_agent,  # post action agent
            unobserved_code=self.unobserved_code,
            agent_indication_raw_value_list = self.agent_indication_raw_value_list,  # called 3rd Nov
            object_count_threshold=object_count_threshold,
            print_out_world_and_plan=print_out_world_and_plan,
        )

        if print_out_world_and_plan and False:
            print_world_at_plan_end(self.time_counter, self.focus_set, pre_action_world=copy.deepcopy(self.internal_preaction_world), post_action_ground_truth_world=copy.deepcopy(self.internal_world), plan=self.whole_plan, pre_action_agent=self.pre_action_agent, active_rules=self.rules)

        for k in stats.keys():
            self.statistics[k] += stats[k]
        self.statistics["observe_call_count"] += 1


    def save_rules_to_file(self, rule_evidence_filename="rule_evidence.json", filename="stepper_state.json"):

        _1, converted_rules = nace.test_utilities.convert_rules_to_char_or_int_mappings(self.rules, self.internal_world, dest="INT")
        _2, converted_negrules = nace.test_utilities.convert_rules_to_char_or_int_mappings(self.negrules, self.internal_world,
                                                                                        dest="INT")
        _3, converted_rule_evidence = nace.test_utilities.convert_rules_to_char_or_int_mappings(self.rule_evidence, self.internal_world,dest="INT")
        _4, converted_rules_excluded = nace.test_utilities.convert_rules_to_char_or_int_mappings(self.rules_excluded, self.internal_world,dest="INT")

        # check each component can be serialed to json
        json.dumps(converted_rules, indent=2)
        json.dumps(converted_negrules, indent=2)
        json.dumps(converted_rule_evidence, indent=2)
        json.dumps(converted_rules_excluded, indent=2)


        _5, converted_rules_char = nace.test_utilities.convert_rules_to_char_or_int_mappings(self.rules, self.internal_world, dest="CHAR")
        _6, converted_negrules_char = nace.test_utilities.convert_rules_to_char_or_int_mappings(self.negrules, self.internal_world,
                                                                                        dest="CHAR")
        _7, converted_rule_evidence_char = nace.test_utilities.convert_rules_to_char_or_int_mappings(self.rule_evidence, self.internal_world,dest="CHAR")
        _8, converted_rules_excluded_char = nace.test_utilities.convert_rules_to_char_or_int_mappings(self.rules_excluded, self.internal_world,dest="CHAR")

        # check each component can be serialed to json
        json.dumps(converted_rules_char, indent=2)
        json.dumps(converted_negrules_char, indent=2)
        json.dumps(converted_rule_evidence_char, indent=2)
        json.dumps(converted_rules_excluded_char, indent=2)


        agent_indication_emdedded_value_list = self.internal_world._convert_raw_to_embedded_values(self.agent_indication_raw_value_list)

        state = {"rules":converted_rules,
                 "negrules":converted_negrules,
                 "rule_evidence":converted_rule_evidence,
                 "rules_excluded":converted_rules_excluded,
                 "time_counter":self.time_counter,
                 "focus_set":self.focus_set,
                 "agent_indication_emdedded_value_list":agent_indication_emdedded_value_list,
                 "world_state":self.internal_world.get_internal_state(),
                 "approx_human_readable":{
                     "rules_char": converted_rules_char,
                     "negrules_char": converted_negrules_char,
                     "rule_evidence_char": converted_rule_evidence_char,
                     "rules_excluded_char": converted_rules_excluded_char,
                 }
                 }

        s = json.dumps(state, indent=2)
        with open(filename, "w") as f:
            f.write(s)

        # rule evidence in machine readable form.
        nace.rule_evidence_utils.save_rule_evidence_to_disk(self.rule_evidence, rule_evidence_filename)

        return s

    def add_rules_and_evidence_from_file_to_memory(self, rule_evidence_filename):
        loaded_rule_evidence = nace.rule_evidence_utils.load_rule_evidence_from_disk(rule_evidence_filename)
        for k in loaded_rule_evidence.keys():
            if k not in self.rule_evidence:
                self.rule_evidence[k] = loaded_rule_evidence[k]
                self.rules.add(k)
            else:
                self.rule_evidence[k] = ( self.rule_evidence[k][0] + loaded_rule_evidence[k][0], self.rule_evidence[k][1] + loaded_rule_evidence[k][1] )



    def save_statistics_to_file(self, filename="data/stepper_statistics.json"):
        s = json.dumps(self.statistics, indent=2)
        with open(filename, "w") as f:
            f.write(s)




# -*- coding: utf-8 -*-

import sys
import os
# insert the parent directory at the start of the sys path
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

# import modeules here, so they can be re-imported elsewhere in unit tests
import nace




import nace

from nace.world_module_numpy import NPWorld
from nace.numpy_utils import convert_string_list_to_cells

def t1_from_string_list():
    str_list = [
        ['oooooooooooo',
         'o          o',
         'o X        o',
         'o          o',
         'o          o',
         'o          o',
         'oooooooooooo'], (),
        ['0,0,0,0,0,0,0,0,0,0,0,0,',
         '0,0,0,0,0,0,0,0,0,0,0,0,',
         '0,0,0,0,0,0,0,0,0,0,0,0,',
         '0,0,0,0,0,0,0,0,0,0,0,0,',
         '0,0,0,0,0,0,0,0,0,0,0,0,',
         '0,0,0,0,0,0,0,0,0,0,0,0,',
         '0,0,0,0,0,0,0,0,0,0,0,0,']]
    # old_world = NPWorld.from_string_list(old_world_str_list)

    view_dist_x=12
    view_dist_y=5
    board_list_of_str = str_list[0]
    cell_shape_rc=(45, 180, 3)
    char_to_rgb_mapping, rgb_to_char_mapping, raw_rgb_array = convert_string_list_to_cells(board_list_of_str, cell_shape_rc=cell_shape_rc)

    assert raw_rgb_array.shape == (315,2160,3)

    # lookup the rgb ndarray value of the agent
    agent_indication_raw_value_list = [char_to_rgb_mapping["X"]]

    # a map can be created with no board
    world = NPWorld(with_observed_time=True, name="from_string()",
                    initial_value=nace.world_module_numpy.UNOBSERVED_BOARD_VALUE,
                    view_dist_x=view_dist_x, view_dist_y=view_dist_y,
                    rgb_to_char_mapping=rgb_to_char_mapping,
                    raw_nptype=type(raw_rgb_array[0][0][0]),
                    raw_cell_shape=cell_shape_rc,
                    agent_indication_raw_value_list=agent_indication_raw_value_list

                    )

    world.multiworld_print([
        {"Caption": f"Blank Expected (no board):",
         "World": world,
         "Color": nace.color_codes.color_code_white_on_black},
    ], agent_indication_raw_value_list=agent_indication_raw_value_list)

    world.update_world_from_ground_truth_NPArray(
        raw_rgb_array,
        update_mode='VIEW',
        wall_code=nace.world_module_numpy.EMBEDDED_WALL_CODE,
        observed_at=float('-inf'),  # -inf == never seen
        agent_indication_raw_value_list=agent_indication_raw_value_list,
        cell_shape_rc = (45, 180, 3)
        )

    world.multiworld_print([
        {"Caption": f"Should have a board:",
         "World": world,
         "Color": nace.color_codes.color_code_white_on_black},
    ], agent_indication_raw_value_list=agent_indication_raw_value_list)


    return world

def t2_from_string_list():

    str_list = \
        ['oooooooooooo',
         'o          o',
         'o X        o',
         'o          o',
         'o          o',
         'o          o',
         'oooooooooooo']

    world, agent_indication_raw_value_list = NPWorld.from_string(str_list, agent_char_code_list=['X'])
    world.multiworld_print([
        {"Caption": f"Should have a board:",
         "World": world,
         "Color": nace.color_codes.color_code_white_on_black},
    ], agent_indication_raw_value_list=agent_indication_raw_value_list)

    location = world.extract_agent_location_raw(agent_indication_raw_value_list)
    print(location)
    assert location == (2,2)


if __name__ == "__main__":
    t1_from_string_list()
    t2_from_string_list()


import nace
import glob
import json
import numpy as np
import gymnasium as gym



def t2_can_find_agent(path ="./data/agents/*.npy"):
    # check the agents location is where it is expected.
    agent_indication_raw_value_list=[]
    fn_list = glob.glob(path)
    for fn in fn_list:
        cell =  np.load(fn)
        agent_indication_raw_value_list.append(cell)

    env = gym.make("CliffWalking-v0",
                   render_mode="rgb_array",
                   )
    env.reset()

    observation, gym_reward, terminated, truncated, info = env.step(0) # 0 == up
    env_rendered = env.render()
    external_ground_truth_npworld = nace.world_module_numpy.NPWorld(with_observed_time=False, name="external ground truth")
    pre_update_agent_location_embedded_space_list, post_update_agent_location_embedded_space_list, modified_count, pre_update_world = external_ground_truth_npworld.update_world_from_ground_truth_NPArray(
        env_rendered,
        update_mode="VIEW",
        cell_shape_rc=(60,60,3),
        add_surrounding_walls=True,
        wall_code=nace.world_module_numpy.EMBEDDED_WALL_CODE,
        agent_indication_raw_value_list=agent_indication_raw_value_list,
        observed_at=1
    )
    locations = external_ground_truth_npworld.extract_agent_location_raw(agent_indication_raw_value_list)
    assert locations == (3,1)

def t1_world_print(path = "./data/agents/*.npy"):
    agent_indication_raw_value_list=[]
    fn_list = glob.glob(path)
    for fn in fn_list:
        cell =  np.load(fn)
        agent_indication_raw_value_list.append(cell)

    env = gym.make("CliffWalking-v0", #"CliffWalking-v0",
                   render_mode="rgb_array",
                   )
    env.reset()

    observation, gym_reward, terminated, truncated, info = env.step(0) # 0 == up
    env_rendered = env.render()
    external_ground_truth_npworld = nace.world_module_numpy.NPWorld(with_observed_time=False, name="external ground truth")
    pre_update_agent_location_embedded_space_list, post_update_agent_location_embedded_space_list, modified_count, pre_update_world = external_ground_truth_npworld.update_world_from_ground_truth_NPArray(
        env_rendered,
        update_mode="VIEW",
        cell_shape_rc=(60, 60, 3),
        add_surrounding_walls=True,
        wall_code=nace.world_module_numpy.EMBEDDED_WALL_CODE,
        agent_indication_raw_value_list=agent_indication_raw_value_list,
        observed_at=1
    )

    external_ground_truth_npworld.multiworld_print([
        {"Caption": f"External:",
         "World": external_ground_truth_npworld,
         "Color": nace.color_codes.color_code_white_on_black},
    ], agent_indication_raw_value_list=agent_indication_raw_value_list)



def t3_from_string_works():
    world_str_list = [
        ['oooooooooooo',
         'of  o      o',
         'o          o',
         'o   oooooooo',
         'o       u  o',
         'o      X   o',
         'oooooooooooo'], (),
        ['25,25,25,23,24,26,26,26,26,26,26,26,',
         '25,25,25,23,24,26,26,26,26,26,26,26,',
         '20,21,22,23,24,26,26,26,26,26,26,26,',
         '20,21,22,23,24,26,26,26,26,26,26,26,',
         '20,21,22,23,24,25,25,25,25,25,25,25,',
         '19,19,19,19,19,19,19,16,15,14,13,12,',
         '18,18,18,18,18,18,18,16,15,14,13,12,']]

    world, agent_indication_raw_value_list = nace.world_module_numpy.NPWorld.from_string(
        world_str_list[0],
        view_dist_x=12,
        view_dist_y=5,
        agent_char_code_list=['X'],
        observed_times_str_list=world_str_list[2]
    )
    world.multiworld_print([{"World": world, "Color": nace.color_codes.color_code_white_on_blue}])

    # check all agent mappings are correct and match
    print(agent_indication_raw_value_list[0].sum())
    assert agent_indication_raw_value_list[0].sum() == 5025090 # ensure we get the same raw agent value each time.
    agent_embedded_value = world.get_val_for_rgb_cell(agent_indication_raw_value_list[0])
    v = world.get_embedded_val_rc(5,7)
    v2 = world.get_val_for_char('X')
    assert v2 == v

    # check bottom right cell has correct time
    assert world.get_time_at_rc(6,11) == 12




def t8_windowsing_function():


    world_str_list = [
        ['oooooooooooo',
         'of  o      o',
         'o   a      o',
         'o   oooooooo',
         'o   b   u  o',
         'o   c  X   o',
         'oooooooooooo'], (),
        ['25,25,25,23,24,26,26,26,26,26,26,26,',
         '25,25,25,23,24,26,26,26,26,26,26,26,',
         '20,21,22,23,24,26,26,26,26,26,26,26,',
         '20,21,22,23,24,26,26,26,26,26,26,26,',
         '20,21,22,23,24,25,25,25,25,25,25,25,',
         '19,19,19,19,19,19,19,16,15,14,13,12,',
         '18,18,18,18,18,18,18,16,15,14,13,12,']]

    world, agent_indication_raw_value_list = nace.world_module_numpy.NPWorld.from_string(
        world_str_list[0],
        agent_char_code_list=['X'],
        observed_times_str_list=world_str_list[2],
    )

    # check points to copy windows correctly
    points_to_copy = world._get_windowed_rc_locations(rc_locations=[(3,3)], whole_screen_raw_width=12, whole_screen_raw_height=7, cell_width=1, cell_height=1)
    assert (0, 0) in points_to_copy
    assert (5, 5) in points_to_copy

    # create e new world, copy from ground truth into it
    internal_world = nace.world_module_numpy.NPWorld(with_observed_time=True, name="self.internal_world", view_dist_x=3,
                                  view_dist_y=2,
                                  )
    internal_world.update_world_from_ground_truth(1, external_ground_truth_world_model=world, rc_locations=[(3,5)], agent_indication_raw_value_list=[])

    internal_world.multiworld_print([{"World": internal_world, "Color": nace.color_codes.color_code_white_on_blue}])

    # check that if we request a line we actually get it
    (line, len) = internal_world.get_board_line(3, color=None, agent_indication_embedded_value_list=['x'], no_color_on_oldest=False)
    assert  line == "..  ooooo..."




def t4_creating_a_new_world_from_string_preserves_char_mappings():
    # debug string and encodings are what was passed in

    input_world_str_list = ['.........AAAAA',
                      'AcbcbcbcbcbcbA',
                      'AbcbcbcbcbcbcA',
                      'APdededededebA',
                      'AgffffffffffhA',
                      'AAAAAAAAAAAAAA']

    world, agent_indication_raw_value_list = nace.world_module_numpy.NPWorld.from_string(
        input_world_str_list,
        view_dist_x=3,
        view_dist_y=2,
        agent_char_code_list=['P'],
        observed_times_str_list=None
    )

    result_board_str_list = world.debug_board.split("\n")

    for i in range (max(len(result_board_str_list), len(input_world_str_list))):
        input_line = ""
        if i < len(input_world_str_list):
            input_line = input_world_str_list[i]

        result_line = ""
        if i < len(result_board_str_list):
            result_line = result_board_str_list[i]

        print(i, input_line.rjust(40), result_line.rjust(40), "SAME" if input_line==result_line else "DIFF")



    assert world.debug_board.split("\n") == input_world_str_list



if __name__ == "__main__":
    t4_creating_a_new_world_from_string_preserves_char_mappings()
    t8_windowsing_function()


    t3_from_string_works()
    t2_can_find_agent("../data/agents/*.npy")
    t1_world_print("../data/agents/*.npy")



import copy
import math
import json
import time
import random
import xxhash
import os.path
import traceback
import collections
from typing import Type, Dict, Tuple, Set, Any, Union, List

import numpy as np
import nace

import nace.color_codes
from nace.agent_module import Agent
# from nace.hypothesis import *
from nace.hypothesis import Hypothesis_BestSelection, Hypothesis_ValidCondition, Hypothesis_Confirmed, \
    Hypothesis_Contradicted, Hypothesis_TruthValue
from nace.prettyprint import prettyprint_all_actions, prettyprint_rule
from nace.world_module_numpy import NPWorld
from nace.test_utilities import convert_rules_to_char_or_int_mappings, convert_rule_evidence_to_char_or_int_mappings


# used in place of randome actions, to ensure we cycle through all actions.
last_action_in_situation = collections.defaultdict(int)


# See nace.py for type and data structure descriptions


def _get_rc_delta_for_action_list(action_list):
    # ignores walls etc. Only use this for testing.
    c = 0
    r = 0
    for action in action_list:
        if action == 'down':
            r += 1
        if action == 'up':
            r -= 1
        if action == 'left':
            c -= 1
        if action == 'right':
            c += 1
    return r, c


def _plan(world: Type[NPWorld],
          active_rules: set,
          actions,  # : Union[List[function], List[NoneType], List[int], List[str]]
          focus_set,
          agent: Type[Agent],
          max_num_actions: int = 12,
          max_queue_length: int = 300,
          custom_goal = None,
          continue_planning_threshold: float = 0.51,  # planning stops in conf drops below this. why 0.5?
          short_curcuit_planning: bool = True,
          shuffle_actions: bool = True,
          brute_force_focus_next_step: bool = False,
          minimum_cell_predict_confidence_threshold: float = 0.0,
          rules_excluded:set = set(),
          circular_action_sequences = [("turn_left", "turn_right"), ("turn_right", "turn_left"), ("turn_right", "turn_right", "turn_right", "turn_right"), ("turn_left", "turn_left", "turn_left", "turn_left")],
          gather_trace_info=False
          ):
    """
    Plan forward searching for situations of highest reward or lowest AIRIS confidence or oldest age.

    We place len(actions) items on the queue for every one we take off, and there is a queue length limit, so
    increasing the number of actions potentially stops us finding the optimal solution.


    Open Question(s):
        Does this routine need to support a no-op? i.e. do not move or do anything?

    Returns:
    - lowest_conf_actions: List of actions leading to the lowest AIRIS confidence
    - lowest_AIRIS_confidence: The lowest AIRIS confidence score (highest reward for exploring)
    - oldest_age_actions: Actions leading to the oldest world state
    - smallest_time: time the oldest world state encountered
    """
    queue = collections.deque([(copy.deepcopy(world), copy.deepcopy(agent.get_values_inc_prefix()), [], 0, "Initialization Value")])  # (world_state, actions, depth, debug_data)
    encountered = {}  # used to short-circuit search
    encountered_rc = {} # used to sort-circuit search
    end_situation_evaluation_count = 0
    evaluation_count_at_which_reward_found = -1
    evaluation_count_at_which_lower_AIRIS_confidence_found = -1
    evaluation_count_at_which_older_found = -1
    actions = copy.deepcopy(actions)
    if shuffle_actions:
        random.shuffle(actions) # shuffle the actions, but keep them in the same random order for this whole function.
    else:
        print("WARN shuffling of action order is disabled.")

    lowest_conf_actions = []
    lowest_AIRIS_confidence = float("inf")  # smaller is better
    lowest_conf_predicted_score_delta = 0.0
    lowest_conf_stopping_reason = "NI"

    max_conf_actions = []

    oldest_age_actions = []
    oldest_age = -1  # if we see an age (time difference) greater than this, we store it.
    oldest_age_predicted_score_delta = 0.0
    oldest_age_stopping_reason = "NI"

    dead_plans = []
    off_map_edge_count = 0

    biggest_predicted_score_delta_actions = []
    biggest_predicted_score_delta = -0.000001
    biggest_predicted_score_delta_stopping_reason = "NI"

    smallest_predicted_score_delta_actions = []
    smallest_predicted_score_delta = 0.000001
    smallest_predicted_score_delta_stopping_reason = "NI"
    predict_next_world_state_call_count = 0
    temp_stop_reason_rc = [[[]]*world.board.shape[1]]*world.board.shape[0]
    if agent.get_score() >= 1:
        pass
    trace_string = ""
    starting_situations_evaluated = 0
    while queue:  # queue size = number of actions * roll out length
        (current_world, current_agent_values, planned_actions, depth, stopping_reason) = queue.popleft()
        starting_situations_evaluated += 1
        if gather_trace_info:
            trace_string += "starting_situations_evaluated=" + str(starting_situations_evaluated) + "\n"

        # need to store agent values on the queue as well, and add the values delta to them.
        if depth >= max_num_actions:
            # print(f"Max depth of {max_num_actions} reached, stopping search.")
            if gather_trace_info:
                t_stopping_reason = "Max depth"
                dead_plan_list = planned_actions
                rc_delta = _get_rc_delta_for_action_list(dead_plan_list)
                dead_plans.append((rc_delta, t_stopping_reason, dead_plan_list))
                trace_string += "  depth >= max_num_actions" + "\n"
            continue # don't add this or children to the queue, i.e. stop exploring from here

        world_state = (current_world.get_board_hashcode(), tuple(current_agent_values[1:]))
        # skip search from here, if not needed via world state hashcodes
        _should_skip_state = world_state in encountered and depth >= encountered[world_state]
        if gather_trace_info:
            trace_string += "  _should_skip_state=" + str(_should_skip_state) + "\n"
        if _should_skip_state:
            # skip states already evaluated (optimisation)
            # print(f"Skipping state")
            if gather_trace_info:
                t_stopping_reason = "Skipping state - been here before"  # i.e. [left, right, left] after a [left] was evaluated.
                dead_plan_list = planned_actions
                rc_delta = _get_rc_delta_for_action_list(dead_plan_list)
                dead_plans.append((rc_delta, t_stopping_reason, dead_plan_list))
                trace_string += "  _should_skip_state" + "\n"
            continue # don't add this or children to the queue, i.e. stop exploring from here

        encountered[world_state] = depth # store that we have been here to prune the search space and avoid loops

        for action in actions:  # check each action in turn from a known state
            new_planned_actions = planned_actions + [action]
            new_planned_actions_rc_delta = _get_rc_delta_for_action_list(new_planned_actions)
            end_rc_location =  agent.get_rc_loc()[0]+new_planned_actions_rc_delta[0], agent.get_rc_loc()[1]+new_planned_actions_rc_delta[1]
            predicted_world, predicted_agent, new_AIRIS_confidence, __age, agent_values_delta, predicted_score_delta = (
                _predict_next_world_state(
                    focus_set,
                    current_world, action, active_rules, agent, custom_goal,
                    brute_force_focus_next_step=brute_force_focus_next_step,
                    minimum_cell_predict_confidence_threshold=minimum_cell_predict_confidence_threshold
                ))
            predict_next_world_state_call_count += 1
            if gather_trace_info:
                trace_string +=  "  new_planned_actions=" + str(prettyprint_all_actions(new_planned_actions)) + "\n" + \
                    "  new_AIRIS_confidence=" + str(new_AIRIS_confidence) + "\n" + \
                    "  new_planned_actions_rc_delta=" + str(new_planned_actions_rc_delta) +"\n" + \
                    "  end_rc_location="+str(end_rc_location) +"\n" + \
                    "  current_world=" +current_world.debug_board +"\n"+ \
                    "  new_world=" + predicted_world.debug_board + "\n"
            if 0 <= end_rc_location[0] < world.board.shape[0] and 0 <= end_rc_location[1] < world.board.shape[1]:
                is_off_map_edge = False
                new_oldest_age = current_world.get_newest_time() - current_world.get_time_at_rc(end_rc_location[0],end_rc_location[1])
            else:
                off_map_edge_count += 1
                is_off_map_edge = True
                new_oldest_age = 0

            # if predicted_score_delta > 0.0:
            #     print("DEBUG REF178 predicted_score_delta > 0.0")
            # if new_AIRIS_confidence < 1.0:
            #     print("DEBUG REF180 new_AIRIS_confidence < 1.0")

            is_circular_action_sequence = False
            for sequence in circular_action_sequences:
                if len(new_planned_actions) >= len(sequence):
                    match = True
                    for i in range(len(sequence)):
                        if sequence[i] != new_planned_actions[-(len(sequence)) + i ]:
                            match = False
                            break
                    if match:
                        is_circular_action_sequence = True
                        break
            if is_circular_action_sequence:
                continue  # no need to continue evaluating (optimisation) this action
                # as the combination of actions redundant, so should be explored elsewhere.

            is_no_board_or_agent_state_change = (predicted_world.get_board_hashcode() == current_world.get_board_hashcode() and
                    sum(agent_values_delta)==0 )

            if predicted_agent.get_terminated() >= 1 and new_AIRIS_confidence >= continue_planning_threshold:
                if gather_trace_info:
                    t_stopping_reason = "agent terminated"
                    dead_plan_list = copy.deepcopy(new_planned_actions)
                    dead_plans.append((new_planned_actions_rc_delta, t_stopping_reason, dead_plan_list))
                    trace_string += "  "+t_stopping_reason + "\n"
                continue  # no need to continue evaluating (optimisation) this action as we just died


            if ((predicted_score_delta < smallest_predicted_score_delta) or
                    (predicted_score_delta == smallest_predicted_score_delta and
                     len(new_planned_actions) < len(smallest_predicted_score_delta_actions) )):
                smallest_predicted_score_delta_actions = copy.deepcopy(new_planned_actions)
                smallest_predicted_score_delta = predicted_score_delta
                smallest_predicted_score_delta_stopping_reason = "smallest score delta"
                # If we know to die on this move, stop exploring
                # if smallest_predicted_score_delta <= -1.0: # TODO magic number, should be passed in or better learnt.
                #     continue  # no need to continue evaluating (optimisation) this action, as we've stored it as current
                    # worse, and would rather not explore beyond

            if predicted_score_delta > biggest_predicted_score_delta or (predicted_score_delta == biggest_predicted_score_delta and len(new_planned_actions) < len(biggest_predicted_score_delta_actions) ):
                biggest_predicted_score_delta_actions = copy.deepcopy(new_planned_actions)
                biggest_predicted_score_delta = predicted_score_delta
                biggest_predicted_score_delta_stopping_reason = "biggest score delta"
                if evaluation_count_at_which_reward_found == -1:
                    evaluation_count_at_which_reward_found = end_situation_evaluation_count

            if new_AIRIS_confidence <= 1.0 and _is_lower_airis_confidence(new_AIRIS_confidence, lowest_AIRIS_confidence, new_planned_actions,
                                          lowest_conf_actions):  # lower or equal confidence and fewer actions
                lowest_conf_actions = new_planned_actions
                lowest_AIRIS_confidence = new_AIRIS_confidence
                lowest_conf_predicted_score_delta = predicted_score_delta
                lowest_conf_stopping_reason = "lower conf"
                if new_AIRIS_confidence < 1.0 and (evaluation_count_at_which_lower_AIRIS_confidence_found == -1):
                    # only recored the first lower than zero time step
                    evaluation_count_at_which_lower_AIRIS_confidence_found = end_situation_evaluation_count

            if _is_max_airis_confidence(new_AIRIS_confidence, new_planned_actions,
                                          max_conf_actions):  # max confidence and more actions
                max_conf_actions = new_planned_actions

            if _is_older_age(new_oldest_age, oldest_age, new_planned_actions, oldest_age_actions):
                oldest_age_actions = new_planned_actions
                oldest_age = new_oldest_age
                oldest_age_predicted_score_delta = predicted_score_delta
                oldest_age_stopping_reason = "is older"
                if evaluation_count_at_which_older_found == -1: # only set once
                    evaluation_count_at_which_older_found = end_situation_evaluation_count

            if world_state not in encountered_rc:
                encountered_rc[world_state] = set()
                if new_planned_actions_rc_delta in encountered_rc[world_state]:
                    if gather_trace_info:
                        dead_plan_list = copy.deepcopy(new_planned_actions)
                        t_stopping_reason = "hash code and xy delta seen before"
                        dead_plans.append((new_planned_actions_rc_delta, t_stopping_reason, dead_plan_list))
                        temp_stop_reason_rc[end_rc_location[0]][end_rc_location[1]].append(t_stopping_reason)
                        temp_stop_reason_rc[end_rc_location[0]][end_rc_location[1]] = list(
                            set(temp_stop_reason_rc[end_rc_location[0]][end_rc_location[1]]))
                        trace_string += "  new_planned_actions_rc_delta in encountered_rc[world_state]" + "\n"
                    continue # no need to continue evaluating (optimisation) this action
                else:
                    encountered_rc[world_state].add(new_planned_actions_rc_delta)

            # if (is_no_board_or_agent_state_change):  # no effect (e.g. hit a wall ) AND no change in score, keys etc
            #     # if we are wrong about this, unit we attempt this and get more info, so this heuristic may be harmful
            #     if gather_trace_info:
            #         dead_plan_list = copy.deepcopy(new_planned_actions)
            #         t_stopping_reason = "action triggered no change (i.e. hit wall)"
            #         dead_plans.append((new_planned_actions_rc_delta, t_stopping_reason, dead_plan_list))
            #         trace_string += "  "+t_stopping_reason + "\n"
            #     continue  # no need to continue evaluating on this path (optimisation)

            if (new_AIRIS_confidence == float("inf")):
                # something bad happens, avoid. This used to be set if reward for step was -ve. however many
                # environments have -ve reward in usual circumstances, ideally we should be agnostic to
                # reward sign or scale.
                # NOTE this code now unlikely/impossible to trigger
                if gather_trace_info:
                    t_stopping_reason = "something bad happens (-ve score)"
                    dead_plan_list = copy.deepcopy(new_planned_actions)
                    dead_plans.append((new_planned_actions_rc_delta, t_stopping_reason, dead_plan_list))
                    trace_string += "  "+t_stopping_reason + "\n"
                continue  # no need to continue evaluating (optimisation) this action

            if new_AIRIS_confidence >= continue_planning_threshold and not is_off_map_edge:
                # calc value deltas together
                agent_values = list(copy.deepcopy(current_agent_values))
                for i, v in enumerate(agent_values_delta):
                    if i < len(agent_values):
                        agent_values[i] += v
                if starting_situations_evaluated < max_queue_length:
                    # add this to the queue to have extra steps added (confidence of 1 means certain, so keep planning)
                    queue.append((predicted_world, agent_values, new_planned_actions, depth + 1, stopping_reason))
                else:
                    if gather_trace_info:
                        dead_plan_list = copy.deepcopy(new_planned_actions)
                        t_stopping_reason = "No longer adding to queue"
                        trace_string += "  " + t_stopping_reason + "\n"
                        dead_plans.append((new_planned_actions_rc_delta, t_stopping_reason, dead_plan_list))

            else: # confidence dropped, or we are off the board do not search further
                # do not add this result to the queue, so it stops being explored further
                if gather_trace_info:
                    dead_plan_list = copy.deepcopy(new_planned_actions)
                    t_stopping_reason = "AIRIS confidence dropped below 1.0"
                    dead_plans.append((new_planned_actions_rc_delta, t_stopping_reason, dead_plan_list))
                    trace_string += "  "+t_stopping_reason + "\n"

            end_situation_evaluation_count += 1
            if end_situation_evaluation_count % 50 == 0 and end_situation_evaluation_count > 0:
                print("DEBUG _plan() end_situation_evaluation_count=",end_situation_evaluation_count,"len(queue)=",len(queue),"starting_situations_evaluated=",starting_situations_evaluated, str(new_planned_actions))
            trace_string += "end_situation_evaluation_count=" + str(end_situation_evaluation_count) + "\n"

            if new_AIRIS_confidence == float("-inf") and predicted_score_delta <= 0.0:
                print("logic error REF342 REF160")

            if short_curcuit_planning:
                if (( evaluation_count_at_which_reward_found > -1 and biggest_predicted_score_delta > 0.0 ) and
                    # as the confidence in dynamics increases search further
                        ((evaluation_count_at_which_lower_AIRIS_confidence_found > -1 and new_AIRIS_confidence < 0.5 and end_situation_evaluation_count > (len(actions)*1) ) or
                    (evaluation_count_at_which_lower_AIRIS_confidence_found > -1 and new_AIRIS_confidence < 0.6 and end_situation_evaluation_count > (len(actions) * 3)) or
                    (evaluation_count_at_which_lower_AIRIS_confidence_found > -1 and new_AIRIS_confidence < 0.7 and end_situation_evaluation_count > (len(actions) * 9)))
                        and
                    (evaluation_count_at_which_older_found > -1 and oldest_age > 1)):
                    print("INFO short circuiting search as 3 of 3 major conditions found.")
                    queue = collections.deque([])
                    break

    print("INFO _plan() len(queue) at completion=", len(queue), "biggest_score_action_length=",len(biggest_predicted_score_delta_actions),
          "starting_situations_evaluated=",starting_situations_evaluated,
          "number of times _predict() called=",predict_next_world_state_call_count
          )

    planning_statistics = {}
    planning_statistics["plan.evaluation_count"] = end_situation_evaluation_count
    planning_statistics["plan.total_items_placed_on_queue"] = starting_situations_evaluated

    max_evaluation_count = max([evaluation_count_at_which_reward_found,evaluation_count_at_which_lower_AIRIS_confidence_found,evaluation_count_at_which_older_found])

    planning_statistics["plan.short_circuit_advantage"] = max(0,(end_situation_evaluation_count-max_evaluation_count))/(end_situation_evaluation_count +1)
    planning_statistics["plan.queue_at_completion"] = len(queue)
    planning_statistics["plan.biggest_score_action_length"] = len(biggest_predicted_score_delta_actions)
    planning_statistics["plan.evaluation_count_at_which_reward_found"] = evaluation_count_at_which_reward_found
    planning_statistics["plan.evaluation_count_at_which_lower_AIRIS_confidence_found"] = evaluation_count_at_which_lower_AIRIS_confidence_found
    planning_statistics["plan.evaluation_count_at_which_older_found"] = evaluation_count_at_which_older_found
    planning_statistics["plan.predict_next_world_state_call_count_during_plan"] = predict_next_world_state_call_count
    planning_statistics["plan.lowest_AIRIS_confidence"] = lowest_AIRIS_confidence

    if gather_trace_info:
        trace_string += json.dumps(planning_statistics)
        with open(os.path.expanduser("./data/temp.txt"),"w") as f:
            f.writelines(trace_string)

    queue = None

    return (lowest_conf_actions, lowest_AIRIS_confidence, lowest_conf_predicted_score_delta, lowest_conf_stopping_reason,
            oldest_age_actions, oldest_age, oldest_age_predicted_score_delta, oldest_age_stopping_reason,
            biggest_predicted_score_delta_actions, biggest_predicted_score_delta, biggest_predicted_score_delta_stopping_reason,
            smallest_predicted_score_delta_actions, smallest_predicted_score_delta, smallest_predicted_score_delta_stopping_reason,
            planning_statistics
            )


def _is_lower_airis_confidence(new_score, best_score, new_actions, best_actions):
    return (new_score < best_score or
            (new_score == best_score and len(new_actions) < len(best_actions)))

def _is_max_airis_confidence(new_score, new_actions, best_actions):
    # 100% confident, and longer
    return (new_score == 1.0 and len(new_actions) > len(best_actions))


def _is_older_age(new_age, oldest_age, new_actions, best_actions):
    """
    Ages are the difference in the time counter. 0 is now, 1 is 1 time step ago, -inf is the start of time.
    @param new_age:
    @param oldest_age:
    @param new_actions:
    @param best_actions:
    @return:
    """
    return (new_age > oldest_age or
            (new_age == oldest_age and len(new_actions) < len(best_actions)))


def _print_score(score):
    """
    Print score value taking its semantics regarding its value range semantics for planning into account

    @param score:
    @return:
    """
    if 0.0 <= score <= 1.0:
        print("certainty:", score)
    else:
        print("desired: True")


def nacev3_get_next_action(
        time_counter,
        focus_set,
        rule_evidence,
        rc_locations,
        internal_world_model: Type[NPWorld],
        rules_in,
        external_ground_truth_world_model: Type[NPWorld],
        print_debug_info,
        stayed_the_same: bool,
        agent: Type[Agent],
        full_action_list: list,  # used during babling
        agent_indication_raw_value_list: list, # embedded value
        max_num_actions: int, # maximum number of actions in plan
        max_queue_length: int # planning queue depth
):
    """
    Determine the next action for the NACE agent based on current observations and rules.

    Steps:
    1. Limit the agent's field of view (partial observability)
    2. Refine rules based on evidence
    3. Plan forward to calculate favored actions and scores

    :param time_counter: Current time
    :param focus_set: Used in subroutine
    :param rule_evidence: Evidence for rules
    :param rc_locations: Tuple(r, c) list of agent's locations (0,0 is top left)
    :param internal_world_model: Agent's view of the world (mutated)
    :param rules_in: Input rules
    :param external_ground_truth_world_model: Current world state (mutated)
    :param  print_debug_info,
    :param stayed_the_same: bool,
    :param agent,
    :return: Tuple(plan, action, rules_excluded, behavior)


    Notes
        The number of state values that the agent stores, and the number of value delta in the rules must match.
    """

    unittest_code_str = ""

    # Step 1: Update agent's field of view
    if external_ground_truth_world_model is not None:
        modified_count, _ = internal_world_model.update_world_from_ground_truth(
            time_counter,
            external_ground_truth_world_model,
            rc_locations=rc_locations,
            agent_indication_raw_value_list=agent_indication_raw_value_list
        )

    # Step 2: Refine rules based on evidence
    active_rules, rules_excluded = Hypothesis_BestSelection(
        rules=copy.deepcopy(rules_in),
        rules_excluded=set(),
        rule_evidence=rule_evidence,
        include_random_rules=stayed_the_same)
    if agent.get_score() >= 1.0:
        pass

    # Step 3: Plan forward
    (lowest_conf_actions, lowest_AIRIS_confidence, lowest_conf_predicted_score_delta, lowest_conf_stopping_reason,
     oldest_age_actions, oldest_age, oldest_age_predicted_score_delta, oldest_age_stopping_reason,
     biggest_predicted_score_delta_actions, biggest_predicted_score_delta, biggest_predicted_score_delta_stopping_reason,
     smallest_predicted_score_delta_actions, smallest_predicted_score_delta, smallest_predicted_score_delta_stopping_reason,
     statistics) = (
        _plan(
            internal_world_model,
            active_rules,
            copy.deepcopy(full_action_list),
            focus_set,
            agent=agent,
            custom_goal=None,
            max_num_actions= max_num_actions,
            max_queue_length= max_queue_length,
            rules_excluded=rules_excluded
        ))

    # Determine available actions
    actions_to_pick_from = _get_remaining_actions(active_rules, full_action_list)
    if len(actions_to_pick_from) == 0:
        all_actions_tried = True
        actions_to_pick_from = copy.deepcopy(full_action_list) # set back to the bigger list so one can be picked at random
    else:
        all_actions_tried = False

    # if len(full_action_list) != 3:
    #     print("login error REF515")

    world_and_agent_hc = xxhash.xxh32(bytes(internal_world_model.board.data) + bytes(agent.get_values_for_precondition())).intdigest()  # should be repeatable between restarts.

    # Determine behavior and action
    behavior, whole_plan = _determine_behavior_and_action(
        lowest_AIRIS_confidence, lowest_conf_actions, lowest_conf_predicted_score_delta, lowest_conf_stopping_reason,
        oldest_age_actions, oldest_age, oldest_age_predicted_score_delta, oldest_age_stopping_reason,
        biggest_predicted_score_delta_actions, biggest_predicted_score_delta,
        biggest_predicted_score_delta_stopping_reason,
        smallest_predicted_score_delta_actions, smallest_predicted_score_delta,
        smallest_predicted_score_delta_stopping_reason,
        all_actions_tried, # True if all possible actions were in the set of applicable rules
        available_actions=actions_to_pick_from,
        full_action_list=full_action_list,
        world_and_agent_hc=world_and_agent_hc,
        print_debug_info=print_debug_info
    )

    if print_debug_info:
        # roll forward the world according to the plan
        predicted_world = copy.deepcopy(internal_world_model)
        planagent = copy.deepcopy(agent)
        for i in range(0, len(whole_plan)):
            predicted_world, planagent, _, __, ___, ____ = _predict_next_world_state(
                focus_set,
                copy.deepcopy(predicted_world), whole_plan[i], active_rules, planagent
            )

        internal_world_model.multiworld_print([{"Caption": f"Internal\nBefore plan:",
                                                "World": internal_world_model,
                                                "Color": nace.color_codes.color_code_white_on_black},
                                               {"Caption": f"At end of plan:\n"+str(prettyprint_all_actions(whole_plan)),
                                                "World": predicted_world,
                                                "Color": nace.color_codes.color_code_white_on_red},
                                               ]
                                              )

        from nace.test_utilities import convert_focus_set_to_char_mapping, convert_rules_to_char_or_int_mappings
        char_focus_set =  convert_focus_set_to_char_mapping(focus_set, internal_world_model)
        char_active_rules, _2 = convert_rules_to_char_or_int_mappings(active_rules, internal_world_model)

        # char_rule_evidence = {}
        # for rule in rule_evidence.keys():
        #     t_char_rules, t_encoded_rules = convert_rules_to_char_or_int_mappings([rule], internal_world_model)
        #     char_rule_evidence[t_encoded_rules[0]] = rule_evidence[rule]

        char_rule_evidence = convert_rule_evidence_to_char_or_int_mappings(rule_evidence, internal_world_model)

        char_world = internal_world_model.debug_board
        unittest_code_str = "world_str_list = "+str(char_world.split("\n")) +"\n"
        unittest_code_str += "focus_set = "+str(char_focus_set)+"\n"
        unittest_code_str += "active_rules = "+str(char_active_rules) + " # subset of rule_evidence \n"
        unittest_code_str += "rule_evidence = "+str(char_rule_evidence) + "\n"

        unittest_code_str += "world_state_values = " +  str(internal_world_model.get_internal_state()) + "\n"

        # (rc_loc=(3,1), score=-696.0, values_exc_score=[])
        unittest_code_str += "agent = nace.agent_module.Agent(rc_loc=" + str(agent.get_rc_loc()) +", score=" + str(agent.get_score()) +", terminated=" + str(agent.get_terminated()) +", values_excluding_prefix=" + str(agent.get_values_exc_prefix()) + " )\n"
        unittest_code_str += "actions = " +str(full_action_list) +"\n"
        unittest_code_str += "behavior_returned = '" + behavior +"'\n" # US spelling no 'u'
        unittest_code_str += "whole_plan_returned = " + str(whole_plan) +"\n"


        # HINT: copy the value of code_str if you want to make a unit test based on the current state.
        print("")

    return whole_plan, rules_excluded, behavior, statistics, unittest_code_str


def _get_remaining_actions(rules, action_list):
    result = copy.deepcopy(action_list)
    for rule in rules:
        precondition, _ = rule
        action = precondition[0]
        if action in result:
            result.remove(action)
    return result


def _determine_behavior_and_action_v0(
                                    # airis_score,
        # favored_actions, favored_actions_for_revisit, oldest_age,
        #                                all_actions_tried: bool, board_value_transition: tuple,
        #                                print_debug_info: bool = False,
                                   lowest_AIRIS_conf,
                                   lowest_AIRIS_conf_actions,
                                   lowest_conf_predicted_score_delta:float,
                                   lowest_AIRIS_conf_stopping_reason: str,
                                   oldest_actions,
                                   oldest_age,
                                   oldest_age_predicted_score_delta:float,
                                   oldest_age_stopping_reason: str,
                                   biggest_predicted_score_delta_actions:list,
                                   biggest_predicted_score_delta:float,
                                   biggest_predicted_score_delta_stopping_reason:str,
                                   smallest_predicted_score_delta_actions:list,
                                   smallest_predicted_score_delta:float,
                                   smallest_predicted_score_delta_stopping_reason:str,
                                   all_actions_tried: bool,
                                   available_actions: list,
                                   full_action_list: list,
                                   world_and_agent_hc,
                                   print_debug_info: bool = False
                                      ):
    """

    Indeterministic if all_actions_tried == FALSE

    @param airis_score: == lowest_AIRIS_conf
    @param favored_actions: lowest_AIRIS_conf_actions or biggest_predicted_score_delta_actions
    @param favored_actions_for_revisit: == oldest_actions
    @param oldest_age: oldest observed age (smaller == older)
    @param all_actions_tried: True if all possible actions were in the set of applicable rules
    @param babbling_rates: dict of rates e.g. : {'curiosity': 0.5, 'exploit': 1.0, 'explore': 0.5}
    @param board_value_transition: tuple ( char, char) from to values of a board transition.
    @param print_debug_info:
    @return:
    """
    plan = []

    favoured_actions = lowest_AIRIS_conf_actions
    # favoured_actions = biggest_predicted_score_delta_actions

    favoured_actions_for_revisit = oldest_actions

    explore_curiosity_modulator = (
        1.0 if all_actions_tried else 0.5
    )  # 1.0 if all_actions_tried and world_is_novel else 0.5
    curiosity_babble_rate, exploit_babble_rate, explore_babble_rate = (
        explore_curiosity_modulator,
        1.0,
        explore_curiosity_modulator,
    )
    exploit_babble = random.random() > (
        exploit_babble_rate
        if lowest_AIRIS_conf == float("-inf")
        else curiosity_babble_rate
    )  # babbling when wanting to achieve something or curious about something, and babbling when exploring:
    explore_babble = (
            random.random() > explore_babble_rate
    )  # since it might not know yet about all ops, exploring then can be limited
    if lowest_AIRIS_conf >= 0.9 or exploit_babble or len(favoured_actions) == 0:
        if (
                not exploit_babble
                and not explore_babble
                and oldest_age > 0.0
                and lowest_AIRIS_conf == 1.0
                and len(favoured_actions_for_revisit) != 0
        ):
            behavior = "EXPLORE"
            # print(
            #     behavior,
            #     Prettyprint_Plan(favoured_actions_for_revisit),
            #     "age:",
            #     oldest_age,
            # )
            action = favoured_actions_for_revisit[0]
            plan = favoured_actions_for_revisit
        else:
            behavior = "BABBLE"
            action = random.choice(available_actions)  # motorbabbling
    else:
        behavior = "ACHIEVE" if lowest_AIRIS_conf == float("-inf") else "CURIOUS"
        # print(behavior, Prettyprint_Plan(favoured_actions), end=" ")
        # NACE_PrintScore(airis_score)
        action = favoured_actions[0]
        plan = favoured_actions

    if print_debug_info:
        print("behavior", behavior,
              "airis_score", lowest_AIRIS_conf,
              "favoured_actions", favoured_actions,
              "favoured_actions_for_revisit", favoured_actions_for_revisit, "oldest_age>0", (oldest_age > 0.0)
              )
    whole_plan = [action] + plan
    return behavior, whole_plan

def _determine_behavior_and_action(lowest_AIRIS_conf,
                                   lowest_AIRIS_conf_actions,
                                   lowest_conf_predicted_score_delta:float,
                                   lowest_AIRIS_conf_stopping_reason: str,
                                   oldest_actions,
                                   oldest_age,
                                   oldest_age_predicted_score_delta:float,
                                   oldest_age_stopping_reason: str,
                                   biggest_predicted_score_delta_actions:list,
                                   biggest_predicted_score_delta:float,
                                   biggest_predicted_score_delta_stopping_reason:str,
                                   smallest_predicted_score_delta_actions:list,
                                   smallest_predicted_score_delta:float,
                                   smallest_predicted_score_delta_stopping_reason:str,
                                   all_actions_tried: bool,
                                   available_actions: list,
                                   full_action_list: list,
                                   world_and_agent_hc,
                                   print_debug_info: bool = False):
    return _determine_behavior_and_action_v2(lowest_AIRIS_conf,
                                   lowest_AIRIS_conf_actions,
                                   lowest_conf_predicted_score_delta,
                                   lowest_AIRIS_conf_stopping_reason,
                                   oldest_actions,
                                   oldest_age,
                                   oldest_age_predicted_score_delta,
                                   oldest_age_stopping_reason,
                                   biggest_predicted_score_delta_actions,
                                   biggest_predicted_score_delta,
                                   biggest_predicted_score_delta_stopping_reason,
                                   smallest_predicted_score_delta_actions,
                                   smallest_predicted_score_delta,
                                   smallest_predicted_score_delta_stopping_reason,
                                   all_actions_tried,
                                   available_actions,
                                   full_action_list,
                                             world_and_agent_hc,
                                   print_debug_info)




def _determine_behavior_and_action_v2(lowest_AIRIS_conf,
                                   lowest_AIRIS_conf_actions,
                                   lowest_conf_predicted_score_delta:float,
                                   lowest_AIRIS_conf_stopping_reason: str,
                                   oldest_actions,
                                   oldest_age,
                                   oldest_age_predicted_score_delta:float,
                                   oldest_age_stopping_reason: str,
                                   biggest_predicted_score_delta_actions:list,
                                   biggest_predicted_score_delta:float,
                                   biggest_predicted_score_delta_stopping_reason:str,
                                   smallest_predicted_score_delta_actions:list,
                                   smallest_predicted_score_delta:float,
                                   smallest_predicted_score_delta_stopping_reason:str,
                                   all_actions_tried: bool,
                                   available_actions: list,
                                   full_action_list: list,
                                   world_and_agent_hc,
                                   print_debug_info: bool = False):
    """

    Nondeterministic if all_actions_tried == FALSE

    Open questions:

    1) Early on in training lowest_AIRIS_conf is high, the system thinks it knows the dynamics fully, yet it doesn't.
    Could this be improved with a familiarity ratting? Is there anything in the literature? We are in an unfamilar situation, proceed with caution?

    2) If a system has a certain amount of stochastic behavior it means the lowest_AIRIS_conf never goes to 1.0, perhaps
    it hovers at 0.33, will be never attempt to achieve?

    3) is it logical to remember what was last done in this situation? Is this over-reliance on heuristics?
    Leading to suboptimal decision-making due to heuristic application?



    EXPLORE  - Refresh old knowledge
    CURIOUS - Go for part of world most uncertain about
    ACHIEVE  - Increase reward
    BABBLE   - Random action


    @param lowest_AIRIS_conf:
    @param lowest_AIRIS_conf_actions:
    @param oldest_actions:
    @param oldest_age: oldest observed age (smaller == older)
    @param all_actions_tried: True if all possible actions were in the set of applicable rules
    @param babbling_rates: dict of rates e.g. : {'curiosity': 0.5, 'exploit': 1.0, 'explore': 0.5}
    @param print_debug_info:
    @return:
    """

    # if all_actions_tried evaluates to False else TRUE randomly 33% of time
    babble_during_explore = False if all_actions_tried else random.random() < 0.33
    explore_during_curious = random.random() < 0.5

    if biggest_predicted_score_delta > 0.0: # TODO: Issue, we will never achieve something with a 0 reward.
        behavior = "ACHIEVE"
        whole_plan = biggest_predicted_score_delta_actions
        action = whole_plan[0]
        reason = "Biggest Score Delta"
    elif lowest_conf_predicted_score_delta > 0.0:
        behavior = "ACHIEVE"
        whole_plan = lowest_AIRIS_conf_actions
        action = whole_plan[0]
        reason = "Score via Lowest AIRIS conf."
    elif oldest_age_predicted_score_delta > 0.0:
        behavior = "EXPLORE"
        whole_plan = oldest_actions
        action = whole_plan[0]
        reason = "Score via oldest age"
    else:
        if (lowest_AIRIS_conf >= 0.90 # we think we model the world well
                or babble_during_explore     # random
                or len(lowest_AIRIS_conf_actions) == 0 # we think we know the world well
        ):
            # EXPLORE or BABBLE
            if (oldest_age > 0.0 and           # we have an old knowledge to recheck
                len(oldest_actions) != 0 and   # we have an old knowledge to recheck
                lowest_AIRIS_conf == 1.0 and   # we think we know the dynamics fully
                not babble_during_explore      # we are not taking a random action
            ):
                behavior = "EXPLORE"
                reason = "OLDEST_CELLS"
                whole_plan = oldest_actions
                action = whole_plan[0]
            else:
                behavior = "BABBLE"
                if len(available_actions) > 0:
                    action = _choose_babble_action(available_actions=available_actions, world_and_agent_hc=world_and_agent_hc)
                else:
                    print("WARN - no actions to pick from, using full list", full_action_list)
                    action = _choose_babble_action(available_actions=full_action_list, world_and_agent_hc=world_and_agent_hc)
                reason = "RANDOM_ACTION"
                whole_plan = [action]
        else:
            # lowest_AIRIS_conf must be < 0.90
            # the oscillation path on world 2 runs through here always selecting lowest_AIRIS_conf_actions
            if explore_during_curious:
                behavior = "EXPLORE"
                reason = "OLDEST_CELLS"
                whole_plan = oldest_actions
                action = whole_plan[0]
            else:
                behavior = "CURIOUS"
                whole_plan = lowest_AIRIS_conf_actions
                action = whole_plan[0]
                reason = "Lowest AIRIS conf."

    if print_debug_info:
        print("DEBUG _determine_behavior_and_action()", "airis_score=", lowest_AIRIS_conf, "lowest_AIRIS_conf_actions=", nace.prettyprint.prettyprint_all_actions(lowest_AIRIS_conf_actions) )
        print("DEBUG _determine_behavior_and_action()", "oldest_age= ", (oldest_age), "oldest_actions=", nace.prettyprint.prettyprint_all_actions(oldest_actions))
        print("DEBUG _determine_behavior_and_action()", "biggest_predicted_score_delta=", biggest_predicted_score_delta,
              "biggest_predicted_score_delta_actions=",
              nace.prettyprint.prettyprint_all_actions(biggest_predicted_score_delta_actions)
              )

    print("INFO _determine_behavior_and_action()", behavior, nace.prettyprint.prettyprint_all_actions([action]),
          "Reason=", reason,
          "Lowest_AIRIS_Conf=", lowest_AIRIS_conf,
          "rest_of_plan=", nace.prettyprint.prettyprint_all_actions(whole_plan),
          "oldest_age=", oldest_age,
          "Lowest_age_reason=", lowest_AIRIS_conf_stopping_reason,
          "oldest_age_stopping_reason=", oldest_age_stopping_reason)
    return behavior, whole_plan


def _choose_babble_action(available_actions, world_and_agent_hc=0, use_random=False):
    if len(available_actions) == 0:
        print("logic error REF 569")

    if use_random:
        return random.choice(available_actions)
    else:
        # be deterministic, i.e. don't do the same as last time, return the next action in the possible actions (will be problematic in continious action spaces)
        # improvement, once the state cache becomes a certain size, switch to random.
        global last_action_in_situation
        index = last_action_in_situation[world_and_agent_hc] % len(available_actions)
        last_action_in_situation[world_and_agent_hc] += 1
        if last_action_in_situation[world_and_agent_hc] > 1:
            print("DEBUG REF842 ")
        return available_actions[index]


def _create_explanation_graphs():
    """
    Hold some things constant, then produce graphs to try an explain how airis score based on synthetic inputs
    e.g.: all_actions_tried, age etc relate to each other
    @return:
    """
    # Calculate babbling rates
    all_results = {}
    num_repeats = 100
    number_airis_bins = 20
    description = ""

    for oldest_age in [10]:
        for all_actions_tried in [True, False]:
            results = []
            for i in range(number_airis_bins + 1):
                airis_score = i / float(number_airis_bins)
                action_type_counts = collections.defaultdict(int)
                behavior_counts = collections.defaultdict(int)

                for run in range(num_repeats):
                    # Determine behavior and action
                    behavior, whole_plan = _determine_behavior_and_action(
                        airis_score,
                        lowest_AIRIS_conf_actions=["FA[0]", "FA[1]"],
                        oldest_actions=["FAFR[0]", "FAFR[1]"],
                        oldest_age=oldest_age,
                        available_actions=["FA[0]", "FA[1]", "FAFR[0]", "FAFR[1]"],
                        full_action_list=["FA[0]", "FA[1]", "FAFR[0]", "FAFR[1]"],
                        all_actions_tried=all_actions_tried,
                    )
                    action = whole_plan[0]
                    behavior_counts[behavior] += 1
                    if action.find("FAFR[") > -1:
                        action_type_counts["favored_actions_for_revisit"] += 1
                    elif action.find("FA[") > -1:
                        action_type_counts["favored_actions"] += 1

                description = "oldest_age:" + str(oldest_age) + " all_actions_tried:" + str(all_actions_tried)
                results.append({"airis_score": airis_score,
                                "CURIOUS": behavior_counts["CURIOUS"],
                                "ACHIEVE": behavior_counts["ACHIEVE"],
                                "BABBLE": behavior_counts["BABBLE"],
                                "EXPLORE": behavior_counts["EXPLORE"],
                                "favored_actions_for_revisit": action_type_counts["favored_actions_for_revisit"],
                                "favored_actions": action_type_counts["favored_actions"],
                                "description": description,
                                'all_actions_tried': all_actions_tried,
                                'oldest_age': oldest_age
                                })
            print("________________ results:")
            print(results)
            all_results[description] = results
    print("________________ all results:")
    print(all_results)
    # return plan, action, rules_excluded, behavior


def _match_hypotheses(focus_set:dict, world:Type[NPWorld], action:any, rules:set, old_agent: Type[Agent], brute_force_focus_next_step:bool):
    """
    Get a map of the grid locations and find the rule with the greatest 'Match Quotient' at each location.
    These are the rules that will actually be applied to the world, and the locations they will be applied in.

    focus_set: seems to only have keys for elements that entered the viewport one at a time,
               i.e. excludes walls or free space.

    This function identifies locations on the grid that are relevant based on
    recent updates and the focus set. It then evaluates how well each rule's
    preconditions match the current state at those locations, assigning a
    "Match Quotient" (Q) for rule r, and cell c. based on the proportion of matched preconditions.

    Note: this routine was called '_MatchHypotheses' in the original code.

    From the paper:
    for each cell (c) we utilize only the rule (r) with Q(r,c) = 1 and maximum fexp(r), meaning the rule preconditions
    not only match perfectly to the given state and the action that has been considered, but also has the
    highest truth expectation among these rule candidates.

    Note: rules are expected to already be filtered for those with the greatest truth expectation

    Problem - there's a q on the map. its position will be empty cause no rules deal with a q. hmmm.


    Args:
        focus_set: A dictionary representing the currently focused elements.
                   Keys are element values, values are counts seen inside viewport, when they appeared 1 at a time.
        world: The current state of the world.
        action: The action to be taken.
        rules: The set of all known rules. Each rule is expected to be a tuple of (precondition, consequence).
               Precondition format: (action, [agent_state_values], (y_rel, x_rel, required_state), ...)
        old_agent: The agent's state in the previous time step.

    Returns:
        A tuple containing:
        - positionscores: A dictionary where keys are (y, x) coordinates and
                         values are tuples of (scores, highscore, highscorerule).
                         'scores' is a dictionary mapping rules to their Match Quotient.
                         'highscore' is the highest Match Quotient at that location.
                         'highscorerule' is the rule with the highest Match Quotient.
        - highesthighscore: The highest Match Quotient found across all locations.


    """


    position_scores = dict([])
    highest_highscore = 0.0
    positions = set([])

    # record locations on the board, whose value is in the focus_set values.
    # Add any surrounding locations from the location deltas in the pre-conditions of any rules.
    height, width = world.get_height_width()
    if brute_force_focus_next_step:
        for y in range(height): # for y,x in rc_locations:
            for x in range(width):
                positions.add((y, x))
    else:
        # attempt to limit locations evaluated to around the focus set
        for y in range(height): # for y,x in rc_locations:
            for x in range(width):
                if world.board[y, x].item() in focus_set:
                    positions.add((y, x))
                    for rule in rules:
                        if action == rule[0][0]:
                            (precondition, consequence) = rule
                            action_and_preconditions = list(precondition)
                            for y_rel, x_rel, requiredstate in action_and_preconditions[
                                                               2:
                                                               ]:
                                positions.add((y + y_rel, x + x_rel))

    # randomised_positions = list(positions)
    # random.shuffle(randomised_positions)


    # find which rules
    for y,x in positions:
        scores = dict([])
        position_scores[(y, x)] = scores
        highscore = 0.0
        highscorerule = None
        rule = None
        for rule in rules:
            (precondition, consequence) = rule
            action_and_preconditions = list(precondition)
            values_excl_score = action_and_preconditions[1] # preconditions exclude the score
            if action_and_preconditions[0] == action:                 # check the ACTION
                scores[rule] = 0.0
            else:
                continue # wrong action - skip to next rule
            if len(values_excl_score) != len(old_agent.get_values_for_precondition()):
                continue  # rule has wrong number of agent values in it - skip to next rule

            CONTINUE = False

            precontions_count = 0
            for i in range(len(values_excl_score)):                             # check the AGENT VALUES against precon
                if len(values_excl_score) != len(old_agent.get_values_for_precondition()):
                    print("WARN Length mis-match REF988 values_excl_score=", values_excl_score,"old_agent.get_values_exc_score()", old_agent.get_values_for_precondition())
                precontions_count += 1
                if  i < len(old_agent.get_values_for_precondition()) and values_excl_score[i] == old_agent.get_values_for_precondition()[
                    i]:  # LHS should not include score, as it is not a state (and should not be) that we condition on)
                    scores[rule] += 1.0  # increment the number of matched preconditions

            for y_rel, x_rel, requiredstate in action_and_preconditions[2:]:# check all precon actually ON BOARD
                if (
                        y + y_rel >= height
                        or y + y_rel < 0
                        or x + x_rel >= width
                        or x + x_rel < 0
                ):
                    CONTINUE = True # if any not on board, skip
                    break
            if CONTINUE:
                continue
            for y_rel, x_rel, requiredstate in action_and_preconditions[
                                               2:]:  # count number of matching preconditions excl action and reward
                    precontions_count += 1
                    if world.board[y + y_rel, x + x_rel] == requiredstate:
                        scores[rule] += 1.0  # increment the number of matched preconditions

            # Q(r,c) - Match Quotient (-2 as it excludes the action and RL reward, but includeds all the agent state values )
            scores[rule] /= precontions_count

            # check the rule has an effect in the 0,0 location or on agent state
            zero_zero_precond = [(y_rel, x_rel, requiredstate) for y_rel, x_rel, requiredstate in
                                 action_and_preconditions[2:] if y_rel == 0 and x_rel == 0]
            has_board_effect = zero_zero_precond[0][2] != consequence[2]
            has_agent_effect = sum([abs(delta) for delta in consequence[3]]) != 0.0

            if scores[rule] > 0.0 and (
                    scores[rule] > highscore
                    # or (
                    #         scores[rule] == highscore
                    #         and highscorerule is not None
                    #         and len(rule[0]) > len(highscorerule[0])
                    # )
                    or (
                            scores[rule] == highscore
                            and (has_board_effect or has_agent_effect)
                            and len(rule[0]) >= len(highscorerule[0])
                    )
            ):
                highscore = scores.get(rule, 0.0)
                highscorerule = rule

        # TODO delete the next block of code - which only creates a human readable version of the rule
        pp_highscore_rule = ""
        if highscorerule is not None:
            try:
                _, char_rules = nace.test_utilities.convert_rules_to_char_or_int_mappings(rules=[highscorerule], world=world)
                char_rule = char_rules[0]
                pp_highscore_rule = nace.prettyprint.prettyprint_rule(rule_evidence={char_rule: (1, 1)},
                                                                      Hypothesis_TruthValue=Hypothesis_TruthValue,
                                                                      rule=char_rule, val_to_char_mappings={},
                                                                      print_evidence=False, print_truth_value=False)
            except Exception as e:
                print("ERROR can not create debugging info."+str(traceback.format_exc()))

        # TODO END of delete code.
        position_scores[(y, x)] = (scores, highscore, highscorerule , pp_highscore_rule)

        if highscore > highest_highscore:
            highest_highscore = highscore
    return (position_scores, highest_highscore)




def _rule_applicable(scores, highscore, highesthighscore, rule):  # called _RuleApplicable in old system
    """
    # Whether a rule is applicable: only if it matches better than not at all, and as well as the best matching rule

    @param scores:
    @param highscore:
    @param highesthighscore:
    @param rule:
    @return:
    """
    if highscore > 0.0 and scores.get(rule, 0.0) == highesthighscore:
        return True
    return False


def _apply_position_match_quotient_scores_to_world_and_agent(
        world,
        agent,
        position_match_quotient_scores,
        highest_match_quotient_highscore,
        action,
        minimum_cell_predict_confidence_threshold,
        custom_goal=None,
    ):
    """
    apply all highest confidence rule for each cell location in that location.
    only rules of the highest confidence will be applied.

    @param world:
    @param agent:
    @param position_match_quotient_scores:
    @param highest_match_quotient_highscore:
    @param action:
    @param minimum_cell_predict_confidence_threshold: only apply a rule to a cell if the confidence is above this value
    @return:
    """
    new_world = copy.deepcopy((world))
    new_agent = copy.deepcopy(agent)
    used_rules_sum_quotient_score = 0
    used_rule_count = 0
    agent_state_increment_counts = [0] * len(new_agent.get_values_inc_prefix())
    debug_rules_applied = []
    for (y, x) in position_match_quotient_scores.keys():
        match_quotient_scores, match_quotient_highscore, rule, pp_rule = position_match_quotient_scores[(y, x)]
        if rule is not None and rule[0][0] == action:
            debug_rule = nace.test_utilities.convert_rules_to_char_or_int_mappings([rule], world) if rule is not None else ""
            if _rule_applicable(match_quotient_scores, match_quotient_highscore, highest_match_quotient_highscore, rule):
                if match_quotient_scores.get(rule, 0.0) > minimum_cell_predict_confidence_threshold:
                    #
                    #
                    # increment the agent state values, and count how many were changed
                    new_agent.accumulate_values_incl_prefix(rule[1][3])  # new agent score should be considered delta (should we call increment_values_including_score?)
                    for i in range(len(rule[1][3])): # count how many times each value is incremented by non zero value
                        if rule[1][3][i] != 0:
                            agent_state_increment_counts[i] += 1
                    # set the value in this cell location
                    new_world.set_embedded_val_rc(y+rule[1][0], x+rule[1][1], rule[1][2], ) # this changes the world value # NOTE effect (r,c) (rule[1][0] and rule[1][1]) are always (0,0) in this impl
                    # gather stats so that confidence can be calculated
                    used_rules_sum_quotient_score += match_quotient_scores.get(rule, 0.0)
                    used_rule_count += 1
                    debug_rules_applied.append( (y, x, debug_rule) )
                    # debugging code - so the map can be viewed. The next line can and should be removed for performnce reasons
                    new_world.debug_board = new_world.board_as_string(
                        agent_indication_embedded_value_list=[] # unknown at this point, rely on enough mapping values being stored
                    )

    new_world.debug_board = new_world.board_as_string(
        agent_indication_embedded_value_list=[] # unknown at this point, rely on enough mapping values being stored
    )

    AIRIS_confidence = (
        used_rules_sum_quotient_score / used_rule_count if used_rule_count > 0 else 1.0 # No rules used, so we have high confidence.
    )

    if (custom_goal and custom_goal(new_world)):
        new_agent.set_score(new_agent.get_score() + 1.0)  # newagent score should be considered as a delta here

    # calculate agent delta values
    agent_value_deltas =  [0]*len(new_agent.get_values_inc_prefix())
    for i in range(len(agent_value_deltas)):
        if agent_state_increment_counts[i] != 0:
            agent_value_deltas[i] = (new_agent.get_values_inc_prefix()[i] - agent.get_values_inc_prefix()[i]) / agent_state_increment_counts[i]
        else:
            agent_value_deltas[i] = (new_agent.get_values_inc_prefix()[i] - agent.get_values_inc_prefix()[
                i])

    return (new_world,
            new_agent,
            AIRIS_confidence,
            None, # unused
            agent_value_deltas,  # delta of all values including score
            agent_value_deltas[0] # score is the first agent value
            )


def _predict_next_world_state(focus_set,
                              world,
                              action,
                              rules,
                              agent,
                              custom_goal=None,
                              brute_force_focus_next_step=False,
                              minimum_cell_predict_confidence_threshold:float=0.0
                              ):  # called NACE_Predict in old system
    """
    Returns the world as predicted after 'action' is applied to 'oldWorld'.
    It does this by applying rules that have been previously learnt.

    How: Apply the move to the predicted world model whereby we use the learned rules to decide how grid elements might
    most likely change.

    From the paper:
    for each cell we utilize only the rule r with Q(r,c) = 1 and maximum fexp(r), meaning the rule preconditions
    not only match perfectly to the given state and the action that has been considered, but also has the
    highest truth expectation among these rule candidates.


    From AIRIS video: https://youtu.be/40W2OmV_rm0?t=371



    :param focus_set: - Note: NOT updated by this routine.
    :param world:
    :param action:
    :param rules:
    :param custom_goal: function that will increase the RL reward score if evalutes to true (takes world as a single param)
    brute_force_next_step: if true evaluates for all cells, not just those derived from focus_set
    minimum_cell_predict_confidence_threshold: only apply a rule to a cell if the confidence is above this value
    :return:
    """

    (position_match_quotient_scores, highest_match_quotient_highscore) = _match_hypotheses(
        focus_set,
        world,
        action,
        rules,
        agent,
        brute_force_focus_next_step=brute_force_focus_next_step # if true use my focus set rather than patricks
    )
    return _apply_position_match_quotient_scores_to_world_and_agent(
        world,
        agent,
        position_match_quotient_scores,
        highest_match_quotient_highscore,
        action,
        minimum_cell_predict_confidence_threshold,
        custom_goal=None)


def _add_to_adjacent_set(adjacent_change_sets: list, newEntry: tuple, MaxCapacity: int, CanCreateNewSet: bool,
                         maximum_distance: int = 1, max_dist_one_dim=True):  # called _AddToAdjacentSet in the old system
    """


    If there are no entries in the adjacent_change_sets create a new one with the new point.
    If the new point if not adjacent to any of the existing sets create a new set containing it.
    If the new point is adjacent to a point in the existing sets, add it to that set if
              there are less than 3 points in it.

    @param adjacent_change_sets:
    @param newEntry:
    @param MaxCapacity:
    @param CanCreateNewSet:
    @param maximum_distance:
    @return:
    """
    (y, x) = newEntry
    AdjacentToAnySet = False
    for consideredSet in adjacent_change_sets:
        consideredSetFrozen = copy.deepcopy(consideredSet)
        for ys, xs in consideredSetFrozen:
            if max_dist_one_dim:
                if abs(y - ys) + abs(x - xs) <= maximum_distance:
                    if len(consideredSet) < MaxCapacity:
                        consideredSet.add(newEntry)
                    AdjacentToAnySet = True
            elif abs(y - ys) <= maximum_distance and abs(x - xs) <= maximum_distance:
                if len(consideredSet) < MaxCapacity:
                    consideredSet.add(newEntry)
                AdjacentToAnySet = True

    if not AdjacentToAnySet and CanCreateNewSet:
        adjacent_change_sets.append({newEntry})


def _is_presently_observed(Time, world, y, x):
    """
    # Whether the grid cell has been observed now (not all have been, due to partial observability)

    NOTE world.get_locations_updated_tminus() can be far more efficient.

    @param Time:
    @param world:
    @param y:
    @param x:
    @return:
    """
    diff = Time - world.times[y][x]
    return diff == 0.0


def _build_change_sets(
        focus_set,
        oldworld,
        action,
        newworld,
        predictedworld,
        object_count_threshold=1,
        agent_rc_location = None
):
    """
    calc value_counts : counts each board value has been seen.

    This routine maintains the 'focus_set' - a list of board values, and the count of times they have changed IFF there
    was only 1 occurrence of the value in the view port.
    in other words: focus set is a set of things that came into the viewport one by one at some stage.

    adjacent_change_sets : sets of 2 to 3 locations that are adjacent to each other in the plane, i.e. {(2,2),(2,3)}
    or {(2,2),(2,1),(2,3)}

    Currently, does not support diagonally adjacent locations.


    @param focus_set:
    @param oldworld:
    @param action:
    @param newworld:
    @param predictedworld:
    @param object_count_threshold:
    @return:
    """
    # Keep track of cell type counts - count of each cell type
    # in world keyed by the chars in the world map
    value_counts = collections.defaultdict(int)
    height, width = oldworld.get_height_width()
    for y in range(height):
        for x in range(width):
            val = oldworld.board[y, x].item()
            value_counts[val] = +1
    # Update focus_set based on unique values and unique changing values.
    # focus_set : keyed by the char in a map location.
    #         Value == the number of time steps this value has been one that has changed (and there was only 1 of them)
    # if the world is huge, is this realistic, or work out from current location till N are found ...
    #
    pass
    # only consider cells that were observed in both newworld and oldworld in the most recent upadate of each,
    # if there is only 1 cell of this type (why?) add this, or increment this in the focus set.
    # Optimisation : Code could be changed to pull this set of locations from the worlds utilizing fast np operations.
    # current_time = newworld.get_newest_time()
    difference_count_checksum =  0

    _a = newworld.get_locations_updated_tminus(0)
    _b = newworld.get_locations_updated_tminus(1)

    recently_observed_rc_locations = _a + _b
    recently_observed_rc_locations = list(set(recently_observed_rc_locations))

    # reduce the set to those changed. This code prevents locations that were unobserved last timestep,
    # but now are (is this intentional) (if not only targets that are now observed should be used?)
    recently_observed_rc_locations = [ (r,c) for (r,c) in recently_observed_rc_locations if oldworld.board[r, c].item() != newworld.board[r, c].item() and oldworld.board[r, c].item() != oldworld.get_unobserved_code() ]

    if len(recently_observed_rc_locations) == 0:
        print("WARN _build_change_sets() No recently updated cells, nothing to learn.")

    for (r,c) in set(recently_observed_rc_locations):
        val = oldworld.board[r, c].item()  # value in the map
        if val not in focus_set:  # if 'val' is new, init focus set
            focus_set[val] = 0
        if val != newworld.board[r, c].item():  # only when it comes into view first time
            if 1 <= value_counts[val] <= object_count_threshold:  # unique - why check for uniqueness?
                # this makes no sense. # dv-5/Aug # Patrick said implementation detail, it meant that it would only
                # trigger if there was 1 new value within visible window
                focus_set[val] += 1  # can happen when already set to 0, or over multiple time steps.
                # because we may focus over multiple time frames?
            difference_count_checksum += 1

    print("DEBUG _build_change_sets() Number of differences between old and new world", difference_count_checksum )

    if agent_rc_location is not None:
        current_agent_locations = [agent_rc_location]
    else:
        current_agent_locations = []


    # create adjacent change sets
    adjacent_change_sets = []
    for (r,c) in recently_observed_rc_locations:
        if (  # if different and observed code
                oldworld.board[r, c].item() != newworld.board[r, c].item()
                and oldworld.board[r, c].item() != oldworld.get_unobserved_code()
        ):
            # mutate changesets adding values to it.
            _add_to_adjacent_set(
                adjacent_change_sets, (r, c), MaxCapacity=3, CanCreateNewSet=True, maximum_distance=1, max_dist_one_dim=False
                # MaxCapacity=3 as 3 things need to have changed for more complex rules to be induced.
                # TODO 3 and 1 are magic numbers.
            )

    changesetslen = len(adjacent_change_sets)  # the length of change set only
    changeset0len = 0  # length of changeset 0 (the first changeset) Possibly this should be the max changeset length?
    if changesetslen > 0:
        changeset0len = len(
            adjacent_change_sets[0]
        )  # temporary fix: 3 things need to have changed at least to allow for the more complex rules to be induced
    # Add prediction mismatch entries to adjacent change set entry (using newworld for observation times)

    for (r,c) in recently_observed_rc_locations:
        if (
                predictedworld
                and predictedworld.board[r, c].item() != newworld.board[r, c].item()
                and oldworld.board[r, c].item() != oldworld.get_unobserved_code()
        ):
            _add_to_adjacent_set(
                adjacent_change_sets, (r, c), MaxCapacity=2, CanCreateNewSet=True, maximum_distance=1
                # MaxCapacity is one less than it was. hmmm...
            )


    # Does the following code fail if (say) the action was right which took effect, but the agent was also
    # forced down?
    # if there was a change next to a focus set element (spatial dependency) add it to the changeSet.
    # this will add
    chgsets = copy.deepcopy(adjacent_change_sets)  #chgsets
    for changeset in chgsets:
        # given a point, and action, generate all the points that could be eligible for use in a rule into a set
        for y, x in changeset:

            # if touching the agent, and the agent location not in change set, add it to the change set.
            for (agent_r,agent_c) in current_agent_locations:
                if math.sqrt(((y-agent_r)*(y-agent_r)) + ((x-agent_c)*(x-agent_c))) <= 1.0:
                    _add_to_adjacent_set(
                        adjacent_change_sets, (agent_r,agent_c), MaxCapacity = 3, CanCreateNewSet = False, maximum_distance = 2
                    )


            if (
                    (action in ['left','right'])
                    and x > 0  # assumes there will be a wall round the board
                    and newworld.board[y, x-1] in focus_set # set of cell values that have changed, but not too many at the same time
                    and oldworld.board[y, x-1] != oldworld.get_unobserved_code()
            ):
                _add_to_adjacent_set(
                    adjacent_change_sets, (y, x - 1), MaxCapacity=3, CanCreateNewSet=False, maximum_distance=1
                )
            if (
                    (action in ['left','right'])
                    and x < width - 1  # assumes there will be a wall round the board
                    and newworld.board[y, x+1] in focus_set
                    and oldworld.board[y, x+1] != oldworld.get_unobserved_code()
            ):
                _add_to_adjacent_set(
                    adjacent_change_sets, (y, x + 1), MaxCapacity=3, CanCreateNewSet=False, maximum_distance=1
                )
            if (
                    (action in ['up','down']) # also had drop in this list
                    and y > 0  # assumes there will be a wall round the board
                    and newworld.board[y-1, x] in focus_set
                    and oldworld.board[y-1, x] != oldworld.get_unobserved_code()
            ):
                _add_to_adjacent_set(
                    adjacent_change_sets, (y - 1, x), MaxCapacity=3, CanCreateNewSet=False, maximum_distance=1
                )
            if (
                    (action  in ['up','down'])
                    and y < height - 1  # assumes there will be a wall round the board
                    and newworld.board[y+1, x] in focus_set
                    and oldworld.board[y+1, x] != oldworld.get_unobserved_code()
            ):
                _add_to_adjacent_set(
                    adjacent_change_sets, (y + 1, x), MaxCapacity=3, CanCreateNewSet=False, maximum_distance=1
                )
            # the next line can add diagonals. (untested)
            # if (
            #         0 < x < width -1  # assumes there will be a wall round the board
            #         and 0 < y < height - 1  # assumes there will be a wall round the board
            # ):
            #     for delta_r, delta_c in [(+1,+1),(-1,-1), (+1,-1), (-1,+1)]:
            #         if (newworld.board[y + delta_r, x+ delta_c] in focus_set
            #         and oldworld.board[y + delta_r, x+ delta_c] != unobserved_code
            #         and (y + delta_r, x+ delta_c) in recently_observed_rc_locations):
            #             _add_to_adjacent_set(
            #                 adjacent_change_sets, (y + delta_r, x+ delta_c), MaxCapacity=3, CanCreateNewSet=True, maximum_distance=1, max_dist_one_dim=False
            #             )



    # Change sets are now built ready to be used.
    print("INFO _build_change_sets() Adjacent change sets have been built:", adjacent_change_sets)
    return focus_set, adjacent_change_sets, changeset0len


def _refine_rules_from_changesets(
        focus_set, # not mutated in this routine or sub routines.
        rule_evidence, # mutated
        oldworld,      # not mutated
        action,        # not mutated
        newworld,      # not mutated
        oldrules,      # copied into new var, mutated and returned
        pre_action_agent,                # not mutated
        ground_truth_post_action_agent,  # not mutated
        predicted_agent,                 # not mutated
        adjacent_change_sets,            # not mutated
        new_negrules,                    # mutated and returned
        changeset0len:int, # ideally we get rid of this parameter. not mutated
                                  ):
    # Build rules based on changes and prediction-observation mismatches
    # Algo: Within each changeset, Compare all changes to all other changes,
    #
    new_rules = copy.deepcopy(oldrules)

    if (predicted_agent.get_values_inc_prefix() != ground_truth_post_action_agent.get_values_inc_prefix()):
        print("DEBUG difference in predicted and actual agent detected. Should be learned in following code.")

    agent_observed_deltas = tuple(  # this stores agent value deltas (inc score)
        [post_v - pre_v for (post_v, pre_v) in (
            zip(ground_truth_post_action_agent.get_values_inc_prefix(),
                pre_action_agent.get_values_inc_prefix()))]  # get the delta of each value inc score
    )

    # STEP 1: Create, and add new rules to new_rules
    for changeset in adjacent_change_sets:
        for y1_abs, x1_abs in changeset:
            action_values_precondition = [action,
                                          tuple(pre_action_agent.get_values_for_precondition())]  # values excluding score
            preconditions = []
            CONTINUE = False
            for y2_abs, x2_abs in changeset:
                (y2_rel, x2_rel) = (y2_abs - y1_abs, x2_abs - x1_abs)  # relative to changeset 1
                condition = (y2_rel, x2_rel, oldworld.board[y2_abs, x2_abs])
                if (
                        oldworld.board[y2_abs, x2_abs] == oldworld.get_unobserved_code()
                ):  # NECESSARY FOR EPISODE RESET ONLY
                    CONTINUE = True  # old world at this location was or now is unobserved. Stop processing this location
                    break
                if Hypothesis_ValidCondition(condition):  # close by (0-2 in distance)
                    preconditions.append(condition)
                    # if current location in old_board has the same object on it in the newworld,
                    # we can skip forward - oct-11 - really? why?
                    if y2_rel == 0 and x2_rel == 0:
                        if (oldworld.board[y2_abs, x2_abs]
                                == newworld.board[y1_abs, x1_abs]):
                            #
                            CONTINUE = True # no change at this location. Stop processing this x,y, location
                            break
            if CONTINUE:
                continue  # skip forward to next location
            preconditions = sorted(preconditions)
            for pr in preconditions:
                action_values_precondition.append(pr)
            rule = (
                tuple(action_values_precondition),
                (
                    0,
                    0,
                    newworld.board[y1_abs, x1_abs],
                    agent_observed_deltas,
                ),
            )
            if len(preconditions) >= 1 and (changeset0len == 3 or len(preconditions) <= 3):
                val_to_char_mappings = newworld.get_val_to_char_mappings()
                rule_evidence, new_rules = Hypothesis_Confirmed(  # Mutates returned copy of rule_evidence and ruleset
                    focus_set, rule_evidence, new_rules, new_negrules, rule, val_to_char_mappings # dv changed to new_rules from oldrules, wed 18th dec 11pm.
                    # note newrules can be mutated in this routine.
                )
        # break  # speedup (dv-this looks odd, why have the outer loop if we always break and only process first element?)
        # The line above means it will only ever process the first changeset


    # STEP 2: create rules for those partialy met, or different outcome
    # if rule conditions are only partly met or the predicted outcome is different from observed,
    # build a specialized rule which has the precondition and conclusion corrected!
    max_focus = None
    if len(focus_set) > 0:
        max_focus = max(focus_set, key=lambda m: focus_set[m])
    (positionscores_v1, highesthighscore_v1) = _match_hypotheses(  # scores == Match Quotient
        focus_set,
        oldworld, action, new_rules, pre_action_agent,
        brute_force_focus_next_step=False
    )

    positionscores =  positionscores_v1
    highesthighscore = highesthighscore_v1

    height, width = oldworld.get_height_width()
    current_time = newworld.get_newest_time()
    val_to_char_mappings = newworld.get_val_to_char_mappings()

    for y in range(height):
        for x in range(width):
            if (y, x) not in positionscores:
                continue
            if (
                    not _is_presently_observed(current_time, newworld, y, x)
                    and oldworld.board[y, x].item() != max_focus
                    and not (
                    newworld.board[y, x].item() == oldworld.get_unobserved_code() and oldworld.board[y, x].item() != oldworld.get_unobserved_code())
            ):
                continue
            scores, highscore, rule, pp_rule = positionscores[(y, x)]
            # for rule in new_rules: # changed from oldrules to new_rules jan 28 (presumeably after copy() was added above) Note: this foreloop was commented out in Patricks code, so we used rule from position scores
            if _rule_applicable(scores, highscore, highesthighscore, rule): # best matching rule
                rule_value_deltas = rule[1][3]
                board_value = newworld.board[y, x]
                if rule[1][2] != board_value or rule_value_deltas != agent_observed_deltas:
                    (precondition, consequence) = rule
                    action_score_and_preconditions = list(precondition)
                    # values = action_score_and_preconditions[1]  # dv commented out as not used 3/aug/2024
                    corrected_preconditions = []
                    CONTINUE = False
                    is_max_focus_value = False
                    for y_rel, x_rel, requiredstate in action_score_and_preconditions[
                                                       2:
                                                       ]:
                        if (
                                y + y_rel >= height
                                or y + y_rel < 0
                                or x + x_rel >= width
                                or x + x_rel < 0
                        ):
                            CONTINUE = True
                            break
                        if oldworld.board[y + y_rel, x + x_rel] == max_focus:
                            is_max_focus_value = True
                        if oldworld.board[y + y_rel, x + x_rel] == oldworld.get_unobserved_code():
                            CONTINUE = True
                            break
                        corrected_preconditions.append(
                            (y_rel, x_rel, oldworld.board[y + y_rel, x + x_rel])
                        )
                    corrected_preconditions = sorted(corrected_preconditions)
                    if CONTINUE or not is_max_focus_value: # do not skip if cell value is max focus
                        continue
                    rule_new = (
                        tuple(
                            [
                                action_score_and_preconditions[0],
                                action_score_and_preconditions[1],
                            ]
                            + corrected_preconditions
                        ),
                        tuple(
                            [
                                rule[1][0],
                                rule[1][1],
                                newworld.board[y, x].item(),
                                tuple( # this takes the diff between last and now,
                                    # Possible ENHANCEMENT:
                                    # adjust the run by how far we are out, last 0.0, now 1.0, predicted was 0.75
                                    # then add 0.25 (now-predicted) to rule location.
                                    [post_v - pre_v for (post_v, pre_v) in (
                                        zip(ground_truth_post_action_agent.get_values_inc_prefix(),
                                            pre_action_agent.get_values_inc_prefix()))]
                                    # get the delta of each value inc score
                                ),
                            ]
                        ),
                    )
                    # print("RULE CORRECTION ", y, x, xy_loc, worldchange);
                    # Prettyprint_rule(rule);
                    # Prettyprint_rule(rule_new)
                    rule_evidence, new_rules = Hypothesis_Confirmed(
                        # Mutates returned copy of rule_evidence and ruleset
                        focus_set, rule_evidence, new_rules, new_negrules, rule_new, val_to_char_mappings
                    )
                    break

    # STEP 3: Add negative evidence for rules which prediction contradicts observation
    # Crisp match: (in a classical AIRIS implementation restricted to deterministic worlds: this part would remove
    # contradicting rules from the rule set and would ensure they can't be re-induced)
    for y in range(height):
        for x in range(width):
            if (
                    not _is_presently_observed(current_time, newworld, y, x)
                    and oldworld.board[y, x].item() != max_focus
                    and not (
                    newworld.board[y, x].item() == newworld.get_unobserved_code() and oldworld.board[y, x].item() != oldworld.get_unobserved_code())
            ):
                continue
            for (
                    rule
            ) in (
                    new_rules # changed from oldrules to new_rules jan 28 (presumeably after copy() was added above)
            ):  # find rules which don't work, and add negative evidence for them (classical AIRIS:
                # remove them and add them to newnegrules)
                (precondition, consequence) = rule
                action_valsExScore_and_preconditions = list(precondition)
                valsExScore = action_valsExScore_and_preconditions[1]
                CONTINUE = False
                if action_valsExScore_and_preconditions[0] != action:  # rule did not apply
                    continue
                for i in range(len(valsExScore)):
                    if (
                            i < len(pre_action_agent.get_values_for_precondition()) and valsExScore[i] != pre_action_agent.get_values_for_precondition()[i]  #
                    ):  # value didn't match, rule did not apply
                        CONTINUE = True
                        break
                for y_rel, x_rel, requiredstate in action_valsExScore_and_preconditions[2:]:
                    if (
                            y + y_rel >= height
                            or y + y_rel < 0
                            or x + x_rel >= width
                            or x + x_rel < 0
                    ):
                        CONTINUE = True
                        break
                    if oldworld.board[y + y_rel, x + x_rel] != requiredstate:
                        CONTINUE = True
                        break
                if CONTINUE:
                    continue
                if rule[1][3][0] != ground_truth_post_action_agent.get_score() - pre_action_agent.get_score():
                    val_to_char_mappings = newworld.get_val_to_char_mappings()
                    rule_evidence, new_rules, new_negrules = Hypothesis_Contradicted(
                        # Mutates returned copy of rule_evidence
                        rule_evidence, new_rules, new_negrules, rule, val_to_char_mappings
                    )  # score increase did not happen
                    continue
                num_values = max(len(rule[1][3]), len(ground_truth_post_action_agent.get_values_inc_prefix()))
                if num_values != len(rule[1][3]):
                    # num states does not match between rule and agent, skip this rule
                    CONTINUE = True
                    break

                for k in range(1, num_values):  # wrong value (not score) prediction (we start at index 1)
                    agent_value_k = ground_truth_post_action_agent.get_values_inc_prefix()[k]
                    rule_value_k = rule[1][3][k]
                    if rule_value_k != agent_value_k:
                        val_to_char_mappings = newworld.get_val_to_char_mappings()
                        rule_evidence, new_rules, new_negrules = Hypothesis_Contradicted(
                            # Mutates returned copy of rule_evidence
                            rule_evidence, new_rules, new_negrules, rule, val_to_char_mappings
                        )
                        CONTINUE = True
                        break
                if CONTINUE:
                    continue
                if rule[1][2] != newworld.board[y, x].item():
                    val_to_char_mappings = newworld.get_val_to_char_mappings()
                    rule_evidence, new_rules, new_negrules = Hypothesis_Contradicted(
                        # Mutates returned copy of rule_evidence
                        rule_evidence, new_rules, new_negrules, rule, val_to_char_mappings
                    )
    return focus_set, rule_evidence, new_rules, new_negrules


def _observe_v2( # new version which splits old into 2 parts for testing purposes
        focus_set,
        rule_evidence,
        oldworld,
        action,
        newworld,
        oldrules,
        oldnegrules,
        predictedworld,
        pre_action_agent,
        ground_truth_post_action_agent,
        predicted_agent,
        unobserved_code,
        object_count_threshold,
):
    """
    Extract new rules from the observations by looking only for observed changes and prediction-observation mismatches

    @param focus_set: Set of focus points : set of cell values that have changed, but not too many at the same time
    @param rule_evidence:  dict[rule:(+ve evidence, -ve evidence)]
    @param oldworld: State of the world before the action
    @param action: Action taken
    @param newworld: State of the world after the action
    @param oldrules: Existing rules
    @param oldnegrules: Existing negative rules
    @param predictedworld: Predicted state of the world after the action
    @param pre_action_agent: Agent state before the action
    @param ground_truth_post_action_agent: Ground truth agent state after the action
    @param predicted_agent: what we predicted the agent state to be
    @param unobserved_code: Code representing unobserved states
    @param object_count_threshold: Threshold for object count
    @return: Updated focus_set, rule_evidence, new_rules, new_negrules
    """
    # new_rules = copy.deepcopy(oldrules)
    # new_negrules = copy.deepcopy(oldnegrules)
    focus_set = copy.deepcopy(focus_set)

    focus_set, adjacent_change_sets, changeset0len = _build_change_sets(
        focus_set,
        oldworld,
        action,
        newworld,
        predictedworld,
        object_count_threshold,
        pre_action_agent.get_rc_loc()
    )


    calculated_focus_set, calculated_rule_evidence, calculated_new_rules, calculated_new_negrules = _refine_rules_from_changesets(
          focus_set,
          rule_evidence,
          oldworld,
          action,
          newworld,
          oldrules,
          pre_action_agent,
          ground_truth_post_action_agent,
          predicted_agent,
          adjacent_change_sets,
          oldnegrules,
          changeset0len
    )

    return calculated_focus_set, calculated_rule_evidence, calculated_new_rules, calculated_new_negrules




def nacev3_predict_and_observe(  # called _predict_and_observe in v1
        time_counter,
        focus_set,
        rule_evidence,
        rc_loc,  # pre action agent location note (0,0) is top left
        pre_action_world,  # pre action internal world model (partially observed) before last_action is applied
        rulesin,  # used in part1 and part2
        negrules,
        last_action,
        rulesExcluded,
        post_action_ground_truth_world,  # i.e. this is the post action world model
        pre_action_agent,  # pre action agent
        ground_truth_post_action_agent,  # post action agent
        unobserved_code,  # value which indicates we can not see the true value of the cell
        agent_indication_raw_value_list:list, # raw value of an agent, i.e. 'x', or the rgb value
        object_count_threshold=1,  # how unique values must be before they are considered. was 1, i.e. unique
        print_out_world_and_plan=True):
    """
    Find difference between the actual world, and the predicted world.

    @param time_counter:
    @param focus_set:
    @param rule_evidence:
    @param rc_loc:
    @param pre_action_world:
    @param rulesin:
    @param negrules:
    @param last_action:
    @param rulesExcluded:
    @param post_action_ground_truth_world:
    @param pre_action_agent
    @param ground_truth_post_action_agent
    @param print_out_world_and_plan:
    @return:
    """
    if print_out_world_and_plan:
        t1 = time.time()
        print("INFO nacev3_predict_and_observe() START")
    simulated_world_original = copy.deepcopy(pre_action_world)
    simulated_world_post_action = copy.deepcopy(pre_action_world)
    rc_location_t_minus_1 = simulated_world_original.extract_agent_location_raw(agent_indication_raw_value_list)
    modified_count, _ = simulated_world_post_action.update_world_from_ground_truth(
        time_counter,
        post_action_ground_truth_world,
        [rc_loc, rc_location_t_minus_1],
        agent_indication_raw_value_list=agent_indication_raw_value_list
    )
    predict_next_world_state_call_count_during_observe = 0

    stayed_the_same = modified_count == 0

    active_rules, rulesExcluded = Hypothesis_BestSelection(
        rules=rulesin,
        rules_excluded=rulesExcluded,
        rule_evidence=rule_evidence,
        include_random_rules=stayed_the_same)  # Step 2 # Mutates rules and rulesExcluded by adding and removing items


    predicted_world, predicted_agent, _, __, values, predicted_score_delta = _predict_next_world_state(
        focus_set,
        copy.deepcopy(simulated_world_original), last_action, active_rules, pre_action_agent
    )
    predict_next_world_state_call_count_during_observe += 1

    #debugging code
    if pre_action_agent.get_score() < ground_truth_post_action_agent.get_score():
        pass # place breakpoint here

    # Extract new rules from the observations by looking only for observed changes and prediction-observation mismatches
    focus_set, rule_evidence, newrules, newnegrules = _observe_v2(
        focus_set,
        rule_evidence,
        simulated_world_original,
        last_action,
        simulated_world_post_action,
        active_rules,
        negrules,
        predicted_world,
        pre_action_agent,
        ground_truth_post_action_agent,
        predicted_agent,
        unobserved_code,
        object_count_threshold,
    )
    used_rules = copy.deepcopy(newrules)
    for rule in rulesExcluded:  # add again so we won't lose them
        newrules.add(rule)

    lastplanworld = copy.deepcopy(simulated_world_original)

    if print_out_world_and_plan:
        t2 = time.time()
        print("INFO nacev3_predict_and_observe() END", "time_taken=",(t2-t1))

    stats = {"predict_next_world_state_call_count_during_observe":predict_next_world_state_call_count_during_observe}

    return (
        used_rules,
        focus_set,
        rule_evidence,
        simulated_world_post_action,
        newrules,
        newnegrules,
        values,
        lastplanworld,
        predicted_world,
        stayed_the_same,
        stats
    )

    # return plan, action, rulesExcluded, behavior, simulated_world_staued_the_same_last_time_new_data_copied_in


def print_world_at_plan_end(time_counter, focus_set, pre_action_world, post_action_ground_truth_world, plan, pre_action_agent, active_rules):
    simulated_world_original = copy.deepcopy(pre_action_world)

    # roll forward the world according to the plan
    predicted_world = copy.deepcopy(simulated_world_original)
    for i in range(0, len(plan)):
        lastplanworld = copy.deepcopy(predicted_world)
        predicted_world, predicted_agent, _, __, ___, ____ = _predict_next_world_state(
            focus_set,
            copy.deepcopy(predicted_world), plan[i], active_rules, pre_action_agent
        )

    print("Learning Phase: 1 time step, ground truth, and end of plan.")
    post_action_ground_truth_world.multiworld_print(
        [
            {"Caption": "Internal model:\ndt=-1",
             "World": simulated_world_original,
             "Color": nace.color_codes.color_code_white_on_blue},

            {"Caption": "Predicted\ndt=0\n==GT?",
             "World": predicted_world,
             "Color": nace.color_codes.color_code_white_on_blue},

            {"Caption": f"Ground Truth \ndt=0\nt={time_counter} beliefs={len(active_rules)}:",
             "World": post_action_ground_truth_world,
             "Color": nace.color_codes.color_code_white_on_black},

            {"Caption": "Plan\nsteps=" + str(len(plan)) + "\n" + str(prettyprint_all_actions(plan)),
             "World": predicted_world,
             "Color": nace.color_codes.color_code_white_on_red},
        ]
    )




if __name__ == "__main__":
    _create_explanation_graphs()
